############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2022 Carl Drougge                       #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

# This runs once per python version the server supports methods for.
# On reload, it is killed and started again.
# When launching a method, it forks and calls the method (without any exec).
# Also contains the function that starts these (new_runners) and the dict
# of running ones {version: Runner}

from __future__ import print_function
from __future__ import division

from importlib import import_module
from types import ModuleType
from traceback import print_exc
import hashlib
import os
import sys
import socket
import signal
import struct
import json
import io
import tarfile
import resource
import gc
import re
from threading import Thread, Lock

archives = {}

def mod2filename(mod):
	if isinstance(mod, ModuleType):
		filename = getattr(mod, '__file__', None)
		if filename and filename[-4:] in ('.pyc', '.pyo',):
			filename = filename[:-1]
		return filename
	else:
		return mod

def get_mod(modname):
	mod = import_module(modname)
	filename = mod2filename(mod)
	prefix = os.path.dirname(filename) + '/'
	return mod, filename, prefix

def path_prefix(paths):
	prefix = os.path.commonprefix(paths)
	# commonprefix works per character
	prefix = prefix.rsplit('/', 1)[0] + '/'
	return prefix

class MsgException(Exception):
	__slots__ = ()

def check_picklable(desc, value):
	from accelerator.compat import pickle
	try:
		pickle.dumps(value, 2)
		return
	except Exception as e:
		msg = str(e)
	raise MsgException('Unpicklable %s: %s' % (desc, msg,))

def load_methods(all_packages, data):
	from accelerator.compat import str_types, iteritems
	from accelerator.extras import DotDict, OptionEnum, OptionEnumValue
	from accelerator.extras import RequiredOption, OptionDefault
	from accelerator import __version__ as ax_version
	res_warnings = []
	res_failed = []
	res_hashes = {}
	res_params = {}
	res_descriptions = {}
	def tar_add(name, data):
		assert name.startswith(dep_prefix)
		info = tarfile.TarInfo()
		info.name = name[len(dep_prefix):]
		info.size = len(data)
		tar_o.addfile(info, io.BytesIO(data))
	all_prefixes = set()
	# This is only used for informational warnings, so failure is not a
	# disaster. And failure is somewhat likely when using several runners
	# using the same-ish python version. (Maybe only on python 2.)
	# There appears to be a race where one python reads an incomplete
	# .pyc file written by the other.
	for package in all_packages:
		try:
			all_prefixes.add(get_mod(package)[2])
		except Exception:
			pass
	for package, key in data:
		modname = '%s.a_%s' % (package, key)
		try:
			mod, mod_filename, prefix = get_mod(modname)
			depend_extra = []
			for dep in getattr(mod, 'depend_extra', ()):
				dep = mod2filename(dep)
				if isinstance(dep, str_types):
					dep = str(dep) # might be unicode on py2
					if not dep.startswith('/'):
						dep = prefix + dep
					depend_extra.append(dep)
				else:
					raise MsgException('Bad depend_extra: %r' % (dep,))
			dep_prefix = os.path.commonprefix(depend_extra + [mod_filename])
			# commonprefix works per character (and commonpath is v3.5+)
			dep_prefix = dep_prefix.rsplit('/', 1)[0] + '/'
			with open(mod_filename, 'rb') as fh:
				src = fh.read()
			tar_fh = io.BytesIO()
			tar_o = tarfile.open(mode='w:gz', fileobj=tar_fh, compresslevel=1)
			tar_add(mod_filename, src)
			h = hashlib.sha1(src)
			hash = int(h.hexdigest(), 16)
			likely_deps = set()
			dep_names = {}
			for k in dir(mod):
				v = getattr(mod, k)
				if isinstance(v, ModuleType):
					filename = mod2filename(v)
					if filename:
						for cand_prefix in all_prefixes:
							if filename.startswith(cand_prefix):
								likely_deps.add(filename)
								dep_names[filename] = v.__name__
								break
			hash_extra = 0
			for dep in depend_extra:
				with open(dep, 'rb') as fh:
					data = fh.read()
				hash_extra ^= int(hashlib.sha1(data).hexdigest(), 16)
				tar_add(dep, data)
			for dep in (likely_deps - set(depend_extra)):
				res_warnings.append('%s.a_%s should probably depend_extra on %s' % (package, key, dep_names[dep],))
			res_hashes[key] = ("%040x" % (hash ^ hash_extra,),)
			res_params[key] = params = DotDict()
			# It would have been nice to be able to use ast.get_source_segment
			def find_source(name):
				bname = name.encode('ascii')
				try:
					start = src.find(b'\n' + bname + b' ')
					if start == -1:
						start = src.index(b'\n' + bname + b'=')
					def find_end(startchar, endchar):
						pos = src.index(startchar, start)
						nest = 0
						in_comment = False
						for pos, c in enumerate(src[pos:], pos):
							if in_comment:
								if c == b'\n'[0]:
									in_comment = False
							elif c == b'#'[0]:
								in_comment = True
							elif c == startchar:
								nest += 1
							elif c == endchar:
								nest -= 1
								if nest == 0:
									return pos
					end = None
					for c in src[start:]:
						if c == b'{'[0]:
							end = find_end(b'{'[0], b'}'[0])
							break
						elif c == b'('[0]:
							end = find_end(b'('[0], b')'[0])
							break
						elif c == b'['[0]:
							end = find_end(b'['[0], b']'[0])
							break
					if not end:
						print('Failed to figure out where %s is in %s' % (name, key,))
						end = start
					return slice(start, end)
				except Exception:
					return slice(0, 0)
			res_descriptions[key] = {
				'text': getattr(mod, 'description', '').strip(),
				'interpreter': sys.executable,
				'interpreter_version': sys.version,
				'accelerator_version': ax_version,
			}
			def fmtopt(v):
				if isinstance(v, type):
					return v.__name__
				elif isinstance(v, dict):
					return '{%s}' % (', '.join('%s: %s' % (fmtopt(k), fmtopt(v)) for k, v in v.items()),)
				elif isinstance(v, list):
					return '[%s]' % (', '.join(fmtopt(v) for v in v),)
				elif isinstance(v, OptionEnum):
					return '{%s}' % (', '.join(sorted(map(str, v._valid))),)
				elif isinstance(v, OptionEnumValue):
					return '%r {%s}' % (v, ', '.join(sorted(map(str, v._valid))),)
				elif isinstance(v, RequiredOption):
					return 'RequiredOption(%s%s)' % (fmtopt(v.value), ', none_ok=True' if v.none_ok else '',)
				elif isinstance(v, OptionDefault):
					if v.default is None:
						return 'OptionDefault(%s)' % (fmtopt(v.value),)
					return 'OptionDefault(%s, default=%s)' % (fmtopt(v.value), fmtopt(v.default),)
				else:
					return repr(v)
			for name, default in (('options', {},), ('datasets', (),), ('jobs', (),),):
				params[name] = d = getattr(mod, name, default)
				if d:
					items = {v[0] if isinstance(v, list) else v for v in params[name]}
					if isinstance(d, dict):
						res_descriptions[key][name] = items = {v: [fmtopt(d[v])] for v in items}
					else:
						res_descriptions[key][name] = items = {v: [] for v in items}
					src_part = src[find_source(name)].decode('utf-8', 'backslashreplace')
					item = None
					spill = []
					prev_item = None
					item_indent = 0
					for orig_line in src_part.split('\n'):
						line = orig_line.strip()
						if not line:
							continue
						itempart = line
						if line.startswith('['):
							itempart = line.split(']')[0][1:]
						if itempart.startswith("'"):
							item = itempart[1:itempart.index("'", 1)]
						elif itempart.startswith('"'):
							item = itempart[1:itempart.index('"', 1)]
						elif not line.startswith('#'):
							item = line.split()[0].split('=')[0]
							if item.startswith('[') and item.endswith(']'):
								item = item[1:-1]
						if item != prev_item:
							if spill and item in items:
								items[item].extend(spill)
							prev_item = item
							item_indent = len(re.match(r'^\s*', orig_line).group(0).expandtabs())
							spill = []
						if '#' in line:
							indent, value = orig_line.split('#', 1)
							value = value.strip()
							if value:
								indent = len(indent.expandtabs())
								if item in items and indent > item_indent:
									items[item].append(value)
								else:
									spill.append(value)
					if spill and item in items:
						items[item].extend(spill)
			for name in ('datasets', 'jobs',):
				d = res_descriptions[key].get(name)
				for item in getattr(mod, name, ()):
					if isinstance(item, list):
						d['[%s]' % (item[0],)] = d.pop(item[0])
			equivalent_hashes = getattr(mod, 'equivalent_hashes', ())
			if equivalent_hashes:
				try:
					assert isinstance(equivalent_hashes, dict)
					assert len(equivalent_hashes) == 1
					k, v = next(iteritems(equivalent_hashes))
					assert isinstance(k, str_types)
					if isinstance(v, str_types):
						v = (v,)
					assert isinstance(v, tuple)
					for vv in v:
						assert isinstance(vv, str_types)
						assert len(vv) == 40
				except AssertionError:
					raise MsgException('Read the docs about equivalent_hashes')
				m = re.search(br'^equivalent_hashes\s*=\s*\{[^}]*\}', src, re.MULTILINE)
				assert m, 'Failed to find equivalent_hashes in ' + mod_filename
				start, end = m.span()
				end -= 1 # to get the same hash as the old way of parsing
				h = hashlib.sha1(src[:start])
				h.update(src[end:])
				verifier = "%040x" % (int(h.hexdigest(), 16) ^ hash_extra,)
				if verifier == k:
					res_hashes[key] += v
				else:
					res_warnings.append('%s.a_%s has equivalent_hashes, but missing verifier %s' % (package, key, verifier,))
			tar_o.close()
			tar_fh.seek(0)
			archives[key] = tar_fh.read()
			check_picklable('options/datasets/jobs', res_params[key])
			check_picklable('description', res_descriptions[key])
		except Exception as e:
			if isinstance(e, MsgException):
				print('%s: %s' % (modname, str(e),))
			else:
				print_exc(file=sys.stderr)
			res_failed.append(modname)
			for d in res_hashes, res_params, res_descriptions:
				d.pop(key, None)
			continue
	return res_warnings, res_failed, res_hashes, res_params, res_descriptions

def launch_start(data):
	from accelerator.launch import run
	from accelerator.compat import PY2
	from accelerator.dispatch import close_fds
	if PY2:
		data = {k: v.encode('utf-8') if isinstance(v, unicode) else v for k, v in data.items()}
	prof_r, prof_w = os.pipe()
	# Disable the GC here, leaving it disabled in the child (the method).
	# The idea is that most methods do not actually benefit from the GC, but
	# they may well be significantly slowed by it.
	# Additionally, as seen in https://bugs.python.org/issue31558
	# the GC sometimes causes considerable extra COW after fork.
	# (If prepare_res is something GC-tracked.)
	# On 3.7+ we also call gc.freeze before forking the analysis processes.
	try:
		gc.disable()
		child = os.fork()
		if not child: # we are the child
			try:
				if not data['debuggable']:
					os.setpgrp() # this pgrp is killed if the job fails
				os.close(prof_r)
				keep = [
					prof_w,
					int(os.getenv('BD_STATUS_FD')),
					int(os.getenv('BD_TERM_FD')),
				]
				close_fds(keep)
				data['prof_fd'] = prof_w
				run(**data)
			finally:
				os._exit(0)
	except Exception:
		os.close(prof_r)
		raise
	finally:
		os.close(prof_w)
		gc.enable()
	return child, prof_r

def respond(cookie, data):
	from accelerator.compat import pickle
	res = pickle.dumps((cookie, data), 2)
	header = struct.pack('<cI', op, len(res))
	with sock_lock:
		sock.sendall(header + res)

def launch_finish(cookie, data):
	status = {'launcher': '[invalid data] (probably killed)'}
	result = None
	try:
		child, prof_r, workdir, jobid, method = data
		arc_name = os.path.join(workdir, jobid, 'method.tar.gz')
		with open(arc_name, 'wb') as fh:
			fh.write(archives[method])
		# We have closed prof_w. When child exits we get eof.
		prof = []
		while True:
			data = os.read(prof_r, 4096)
			if not data:
				break
			prof.append(data)
		try:
			status, result = json.loads(b''.join(prof).decode('utf-8'))
		except Exception:
			pass
	finally:
		os.close(prof_r)
		respond(cookie, (status, result))

# because a .recvall method is clearly too much to hope for
# (MSG_WAITALL doesn't really sound like the same thing to me)
def recvall(sock, z, fatal=False):
	data = []
	while z:
		tmp = sock.recv(z)
		if not tmp:
			if fatal:
				sys.exit(0)
			return
		data.append(tmp)
		z -= len(tmp)
	return b''.join(data)

class Runner(object):
	def __init__(self, pid, sock, python):
		self.pid = pid
		self.sock = sock
		self.python = python
		self.cookie = 0
		self._waiters = {}
		self._lock = Lock()
		self._thread = Thread(
			target=self._receiver,
			name="%d receiver" % (pid,),
		)
		self._thread.daemon = True
		self._thread.start()

	# runs on it's own thread (in the server), one per Runner object
	def _receiver(self):
		from accelerator.compat import QueueFull, pickle, itervalues
		while True:
			try:
				hdr = recvall(self.sock, 5)
				if not hdr:
					break
				op, length = struct.unpack('<cI', hdr)
				data = recvall(self.sock, length)
				cookie, data = pickle.loads(data)
				q = self._waiters.pop(cookie)
				q.put(data)
			except Exception:
				print_exc(file=sys.stderr)
				break
		# All is lost, unblock anyone waiting
		for q in itervalues(self._waiters):
			try:
				q.put(None, block=False)
			except QueueFull:
				pass

	def kill(self):
		try:
			self.sock.close()
		except Exception:
			pass
		try:
			os.kill(self.pid, signal.SIGKILL)
		except Exception:
			pass
		try:
			os.waitpid(self.pid, 0)
		except Exception:
			pass

	def _waiter(self, cookie):
		from accelerator.compat import Queue
		q = Queue(1)
		self._waiters[cookie] = q
		return q.get

	# this is called from request threads, so needs locking to avoid races
	def _do(self, op, data):
		from accelerator.compat import pickle
		with self._lock:
			cookie = self.cookie
			self.cookie += 1
			# have to register waiter before we send packet (to avoid a race)
			waiter = self._waiter(cookie)
			data = pickle.dumps((cookie, data), 2)
			header = struct.pack('<cI', op, len(data))
			self.sock.sendall(header + data)
		# must wait without the lock, otherwise all this threading gets us nothing.
		res = waiter()
		if res is None:
			from accelerator.error import AcceleratorError
			raise AcceleratorError("Runner exited unexpectedly.")
		return res

	def load_methods(self, all_packages, data):
		return self._do(b'm', (all_packages, data))

	def launch_start(self, data):
		return self._do(b's', data)

	def launch_finish(self, child, prof_r, workdir, jobid, method):
		return self._do(b'f', (child, prof_r, workdir, jobid, method))

	def launch_waitpid(self, child):
		return self._do(b'w', child)

	def get_ax_version(self):
		return self._do(b'v', None)

runners = {}
def new_runners(config, used_versions):
	from accelerator.dispatch import run
	from accelerator.compat import itervalues, iteritems
	killed = set()
	for runner in itervalues(runners):
		if id(runner) not in killed:
			runner.kill()
			killed.add(id(runner))
	runners.clear()
	candidates = {'DEFAULT': sys.executable}
	for cnt in (1, 2, 3):
		candidates['.'.join(map(str, sys.version_info[:cnt]))] = sys.executable
	candidates.update(config.interpreters)
	todo = {k: v for k, v in candidates.items() if k in used_versions}
	exe2r = {}
	for k, py_exe in iteritems(todo):
		if py_exe in exe2r:
			runners[k] = exe2r[py_exe]
		else:
			sock_p, sock_c = socket.socketpair(socket.AF_UNIX, socket.SOCK_STREAM)
			runner_fn = __file__
			if runner_fn[-4:] in ('.pyc', '.pyo',):
				# workaround for python 2 server running python 3 methods
				runner_fn = runner_fn[:-1]
			cmd = [py_exe, runner_fn, str(sock_c.fileno()), sys.path[0]]
			pid = run(cmd, [sock_p.fileno()], [sock_c.fileno()], no_stdin=not config.debuggable)
			sock_c.close()
			runners[k] = Runner(pid=pid, sock=sock_p, python=py_exe)
			exe2r[py_exe] = runners[k]
	return runners

if __name__ == "__main__":
	# sys.path needs to contain the project dir, but not the accelerator dir
	sys.path[0] = sys.argv[2]

	# As of python 3.8 the default start_method is 'spawn' on macOS.
	# This doesn't work for us. 'fork' is fairly unsafe on macOS,
	# but it's better than not working at all. See
	# https://bugs.python.org/issue33725
	# for more information.
	import multiprocessing
	if hasattr(multiprocessing, 'set_start_method'):
		multiprocessing.set_start_method('fork')

	from accelerator import __version__ as ax_version

	from accelerator.autoflush import AutoFlush
	from accelerator.compat import pickle
	from accelerator.dispatch import update_valid_fds

	sys.stdout = AutoFlush(sys.stdout)
	sys.stderr = AutoFlush(sys.stderr)
	sock = socket.fromfd(int(sys.argv[1]), socket.AF_UNIX, socket.SOCK_STREAM)
	sock_lock = Lock()
	update_valid_fds()

	# Set the highest open file limit we can.
	# At least OS X seems to like claiming no limit as max without
	# allowing that to be set, so let's do some retrying.
	r1, r2 = resource.getrlimit(resource.RLIMIT_NOFILE)
	limits = [500000, 100000, 50000, 10000, 5000, 1000, r1, r2]
	limits = sorted((v for v in limits if r1 <= v <= r2), reverse=True)
	for try_limit in limits:
		try:
			resource.setrlimit(resource.RLIMIT_NOFILE, (try_limit, r2))
			break
		except ValueError:
			pass
	r1, r2 = resource.getrlimit(resource.RLIMIT_NOFILE)
	if r1 < r2:
		print("WARNING: Failed to raise RLIMIT_NOFILE to %d. Set to %d." % (r2, r1,))
	if r1 < 5000:
		print("WARNING: RLIMIT_NOFILE is %d, that's not much." % (r1,))

	while True:
		op, length = struct.unpack('<cI', recvall(sock, 5, True))
		data = recvall(sock, length, True)
		cookie, data = pickle.loads(data)
		if op == b'm':
			res = load_methods(*data)
			respond(cookie, res)
		elif op == b's':
			res = launch_start(data)
			respond(cookie, res)
		elif op == b'f':
			# waits until job is done, so must run on a separate thread
			Thread(
				target=launch_finish,
				args=(cookie, data,),
				name=data[3], # jobid
			).start()
		elif op == b'w':
			# It would be nice to be able to just ignore children
			# (set SIGCHLD to SIG_IGN), but the server might want to
			# killpg the child, so we need it to stick around.
			os.waitpid(data, 0)
			respond(cookie, None)
		elif op == b'v':
			respond(cookie, ax_version)



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2021 Anders Berkeman                    #
# Modifications copyright (c) 2019-2022 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division

import re
import os
import shlex

from accelerator.compat import quote_plus, open

from accelerator.extras import DotDict


_re_var = re.compile(r'(?<!\\)\$\{([^\}=]*)(?:=([^\}]*))?\}')
def interpolate(s):
	"""Replace ${FOO=BAR} with os.environ.get('FOO', 'BAR')
	(just ${FOO} is of course also supported, but not $FOO)"""
	return _re_var.subn(lambda m: os.environ.get(m.group(1), m.group(2)), s)[0]


class HostPortTuple(tuple):
	__slots__ = ()
	def __str__(self):
		return "http://%s:%d" % self

def resolve_listen(listen):
	if listen.startswith('http://'):
		tmp = listen[7:].rstrip('/')
		if '/' not in tmp and ':' in tmp:
			listen = tmp
	if '/' not in listen and ':' in listen:
		hostname, port = listen.rsplit(':', 1)
		listen = HostPortTuple((hostname or 'localhost', int(port),))
		url = 'http://%s:%d' % listen
	else:
		url = None
	return listen, url

def fixup_listen(project_directory, listen):
	listen, url = listen
	if not isinstance(listen, tuple):
		socket = os.path.join(project_directory, listen)
		url = 'unixhttp://' + quote_plus(os.path.realpath(socket))
	return listen, url


def load_config(filename):
	from accelerator.error import UserError

	multivalued = {'workdirs', 'method packages', 'interpreters'}
	required = {'slices', 'workdirs', 'method packages'}
	known = {'target workdir', 'listen', 'urd', 'board listen', 'result directory', 'input directory', 'project directory'} | required | multivalued
	cfg = {key: [] for key in multivalued}

	def fixpath(fn, realpath=True):
		# convert relative path to absolute wrt location of config file
		p = os.path.join(project_directory, fn)
		if realpath:
			p = os.path.realpath(p)
		else:
			p = os.path.normpath(p)
		return p

	class _E(Exception):
		__slots__ = ()
	def parse_package(val):
		if len(val) == 2:
			if val[1] != 'auto-discover':
				raise _E('Either no option or "auto-discover" for package %r.' % (val[0],))
			else:
				return (val[0], True)
		return (val[0], False)
	def parse_workdir(val):
		return val[0], fixpath(val[1])
	def parse_interpreter(val):
		return val[0], fixpath(val[1], False)
	def check_interpreter(val):
		if val[0] == 'DEFAULT':
			raise _E("Don't override DEFAULT interpreter")
		if not os.path.isfile(val[1]):
			raise _E('%r does not exist' % (val,))
	def check_workdirs(val):
		name, path = val
		if name in (v[0] for v in cfg['workdirs']):
			raise _E('Workdir %s redefined' % (name,))
		if path in (v[1] for v in cfg['workdirs']):
			raise _E('Workdir path %r re-used' % (path,))

	def resolve_urd(val):
		orig_val = val
		is_local = (val[0] == 'local')
		if val[0] in ('local', 'remote'):
			if len(val) == 1:
				val = ['.socket.dir/urd']
			else:
				val = [val[1]]
		if len(val) != 1:
			raise _E("urd takes 1 or 2 values (expected %s, got %r)" % (' '.join(parsers['urd'][0]), orig_val))
		return is_local, resolve_listen(val[0])

	parsers = {
		'slices': (['count'], int),
		'workdirs': (['name', 'path'], parse_workdir),
		'interpreters': (['name', 'path'], parse_interpreter),
		'listen': (['path or [host]:port'], resolve_listen),
		'urd': (['["local" or "remote"] and/or', '[path or [host]:port]'], resolve_urd), # special cased
		'board listen': (['path or [host]:port'], resolve_listen),
		'input directory': (['path'], fixpath),
		'result directory': (['path'], fixpath),
		'method packages': (['package', '[auto-discover]'], parse_package),
	}
	checkers = dict(
		interpreter=check_interpreter,
		workdirs=check_workdirs,
	)

	with open(filename, 'r', encoding='utf-8') as fh:
		lines = list(enumerate(fh, 1))
	def parse(handle):
		key = None
		for n, line in lines:
			lineno[0] = n
			line_stripped = line.strip()
			if not line_stripped or line_stripped[0] == '#':
				continue
			if line == line.lstrip():
				if ':' not in line:
					raise _E('Expected a ":"')
				key, val = line.split(':', 1)
				if key not in known:
					raise _E('Unknown key %r' % (key,))
			else:
				if not key:
					raise _E('First line indented')
				val = line
			val = shlex.split(interpolate(val), posix=True, comments=True)
			if val:
				handle(key, val)
	def just_project_directory(key, val):
		if key == 'project directory':
			if len(val) != 1:
				raise _E("%s takes a single value path (maybe you meant to quote it?)" % (key,))
			project_directory[0] = val[0]
	def everything(key, val):
		if key in parsers:
			args, p = parsers[key]
			if key in ('urd', 'method packages'):
				want_count = (1, 2)
				want_count_str = "1 or 2"
			else:
				want_count = [len(args)]
				want_count_str = str(want_count[0])
			if len(val) not in want_count:
				if len(args) == 1:
					raise _E("%s takes a single value %s (maybe you meant to quote it?)" % (key, args[0]))
				else:
					raise _E("%s takes %s values (expected %s, got %r)" % (key, want_count_str, ' '.join(args), val))
			if len(args) == 1:
				val = val[0]
			val = p(val)
		elif len(val) == 1:
			val = val[0]
		else:
			raise _E("%s takes a single value (maybe you meant to quote it?)" % (key,))
		if key in checkers:
			checkers[key](val)
		if key in multivalued:
			cfg[key].append(val)
		else:
			if key in cfg:
				raise _E("%r doesn't take multiple values" % (key,))
			cfg[key] = val

	try:
		project_directory = [os.path.dirname(filename)]
		lineno = [None]
		parse(just_project_directory)
		lineno = [None]
		project_directory = os.path.realpath(project_directory[0])
		parse(everything)
		lineno = [None]

		missing = set()
		for req in required:
			if not cfg[req]:
				missing.add(req)
		if missing:
			raise _E('Missing required keys %r' % (missing,))

		# Reformat result a bit so the new format doesn't require code changes all over the place.
		rename = {
			'target workdir': 'target_workdir',
			'method packages': 'method_directories',
			'input directory': 'input_directory',
			'result directory': 'result_directory',
			'project directory': 'project_directory',
			'board listen': 'board_listen',
		}
		res = DotDict({rename.get(k, k): v for k, v in cfg.items()})
		if 'listen' not in res:
			res.listen = '.socket.dir/server', None
		if 'target_workdir' not in res:
			res.target_workdir = res.workdirs[0][0]
		if 'project_directory' not in res:
			res.project_directory = os.path.dirname(filename)
		res.project_directory = os.path.realpath(res.project_directory)
		res.workdirs = dict(res.workdirs)
		if res.target_workdir not in res.workdirs:
			raise _E('target workdir %r not in defined workdirs %r' % (res.target_workdir, set(res.workdirs),))
		res.interpreters = dict(res.interpreters)
		for exe in res.interpreters.values():
			assert os.path.exists(exe), 'Executable %r does not exist.' % (exe,)
		res.listen, res.url = fixup_listen(res.project_directory, res.listen)
		if res.get('urd'):
			res.urd_local, listen = res.urd
			res.urd_listen, res.urd = fixup_listen(res.project_directory, listen)
		else:
			res.urd_local, res.urd_listen, res.urd = False, None, None
		res.board_listen, _ = fixup_listen(res.project_directory, res.get('board_listen', ('.socket.dir/board', None)))
		res.method_directories = dict(res.method_directories)
	except _E as e:
		if lineno[0] is None:
			prefix = 'Error in %s:\n' % (filename,)
		else:
			prefix = 'Error on line %d of %s:\n' % (lineno[0], filename,)
		raise UserError(prefix + e.args[0])

	res.config_filename = os.path.realpath(filename)
	return res



# This isn't intended as an example, it's just here to simplify all the
# output formating the examples do.

from accelerator import colour

class Printer:
	def __init__(self, indent=0):
		self._indent = indent
		self._last_indent = 0
		self._unindent = []

	def _prefix(self, txt):
		for c in txt:
			if not c.isspace():
				break
			yield c
		yield None

	def _print(self, a, indent, *attrs):
		txt = ' '.join(str(v) for v in a)
		prefixes = []
		lines = txt.lstrip('\n').rstrip().split('\n')
		for line in lines:
			if line.strip():
				prefixes.append(list(self._prefix(line)))
		prefix_len = 0
		if prefixes:
			prefixes[0].pop()
			while len(prefixes[0]) > prefix_len and len(set(p[prefix_len] for p in prefixes)) == 1:
				prefix_len += 1
		lines = [line[prefix_len:].rstrip() for line in lines]
		self._last_indent = max(indent + self._indent, 0)
		indent = ' ' * self._last_indent
		lines = [indent + line if line else '' for line in lines]
		txt = '\n'.join(lines)
		if attrs and txt:
			txt = colour(txt, *attrs)
		print(txt)
		return self

	def __call__(self, *a):
		return self._print(a, 0, 'brightblue')

	def header(self, *a):
		return self._print(a, 0, 'bold', 'brightblue')

	def command(self, *a):
		return self._print(a, 2, 'bold')

	def output(self, *a):
		return self._print(a, 2)

	def plain(self, *a):
		return self._print(a, 0)

	def source(self, filename):
		return self._print(('Source: ' + colour.bold(filename),), -1000)

	def __enter__(self):
		self._unindent.append(self._indent)
		self._indent = self._last_indent + 2
		return self

	def __exit__(self, e_type, e_value, e_tb):
		self._indent = self._unindent.pop()

prt = Printer()



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify that lists of chains are handled in csvexport, including changing
column types.
'''

from accelerator import subjobs

def verify(want, **kw):
	job = subjobs.build('csvexport', labelsonfirstline=False, **kw)
	with job.open('result.csv') as fh:
		got = fh.read()
	assert want == got, 'wanted %r, got %r' % (want, got,)
	return job

def synthesis(job):
	dw = job.datasetwriter(name='a')
	dw.add('fixed', 'ascii')
	dw.add('changing', 'ascii')
	dw.get_split_write()('a', 'a')
	a = dw.finish()
	dw = job.datasetwriter(name='b', previous=a)
	dw.add('fixed', 'ascii')
	dw.add('changing', 'bytes')
	dw.get_split_write()('b', b'\xc3\xa5')
	b = dw.finish()
	dw = job.datasetwriter(name='c', previous=b)
	dw.add('fixed', 'ascii')
	dw.add('changing', 'json')
	dw.get_split_write()('c', None)
	c = dw.finish()

	# first alone just to see that later failures are not something trivial
	verify('a,a\n', source=a)
	verify('\xe5,b\n', source=b)
	verify('null,c\n', source=c)

	verify('a\nb\nc\n', labels=['fixed'], source=[a, b, c])
	verify('a\na\nb\nc\n', labels=['fixed'], source=[a, c], chain_source=True)
	verify('a\nb\nc\n', labels=['fixed'], source=c, chain_source=True)
	verify('a\na\n\xe5\nnull\n', labels=['changing'], source=[a, c], chain_source=True)
	ab = verify('a\n\xe5\n', labels=['changing'], source=[a, b])
	verify('\xe5\nnull\nnull\n', labels=['changing'], source=[c, c], chain_source=True, previous=ab)



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Tests the parsed:type writers.
With plain values, parsable values and unparsable values.
'''

# this is just to make nan values compare equal, but the easiest way
# to do that is to just stringify everything.
# since we aren't comparing different types, this is safe.
def nanfix(lst):
	return [str(v) for v in lst]

def hashfilter(typ, values, sliceno):
	from accelerator.g import slices
	if typ == 'json':
		return values[sliceno::slices]
	else:
		from accelerator.dsutil import typed_writer
		h = typed_writer(typ).hash
		return [v for v in values if h(v) % slices == sliceno]

def synthesis(job, slices):
	def test(typ, write_values, want_values, bad_values=[], allow_none=True):
		dw = job.datasetwriter(
			name=typ,
			columns={'value': ('parsed:' + typ, allow_none)},
			hashlabel=None if typ == 'json' else 'value',
		)
		write = dw.get_split_write()
		for value in write_values:
			try:
				write(value)
			except Exception as e:
				raise Exception('Failed to write %r to %s column: %r' % (value, typ, e))
		for value in ['foo', '1 two'] + bad_values:
			try:
				write(value)
				raise Exception('writer for parsed:%s allowed %r' % (typ, value,))
			except (ValueError, OverflowError):
				pass
		ds = dw.finish()
		for sliceno in range(slices):
			want_slice = hashfilter(typ, want_values, sliceno)
			got_slice = list(ds.iterate(sliceno, 'value'))
			assert nanfix(got_slice) == nanfix(want_slice), "%s got %r, wanted %r in slice %d" % (typ, got_slice, want_slice, sliceno)

	inf = float('inf')
	nan = float('nan')

	test('number',
		['42.0', '42', inf, '-inf', None, '9'*200,  'nan', nan, ' 1e55 ', '+1'],
		[ 42.0,   42,  inf,  -inf,  None, 10**200-1, nan,  nan,   1e55,     1],
		['1+1', '+', '-', ''],
	)
	test('complex32',
		[42-0.5j, '0.5+infj',        None, 'j', '9',  '-inf+nanj',       '+1'],
		[42-0.5j, complex(0.5, inf), None,  1j,  9+0j, complex(-inf, nan), 1+0j],
		['1+1', '+', '-', ''],
	)
	test('complex64',
		[-0.99j, 'inf-1j',         None, '  nan-infj ',      'nan+nanj',        '-j'],
		[-0.99j, complex(inf, -1), None, complex(nan, -inf), complex(nan, nan), 0-1j],
		['1-1', '+', '-', ''],
	)
	test('float32',
		[-0.75, '0.5', '2e3',   None, '1e100', 'nan', nan, '\n\t+5.\r', '011'],
		[-0.75,  0.5,   2000.0, None,  inf,     nan,  nan,       5.0,     11.0],
		['0x1.5p+5', '1+1', '+', '-', ''],
	)
	test('float64',
		[-0.99, '-0.1', None, '1e100', '-inf', 'nan', inf, '1e+2', ' -011 \n'],
		[-0.99,  -0.1,  None,  1e100,   -inf,   nan,  inf,  100.0,   -11.0],
		['0x5', '1+1', '+', '-', ''],
	)
	test('int32',
		[None, '42', 0, '-99', '1', 2, 3, 4, 5, ' 9 ', '011', '-0'],
		[None,  42,  0,  -99,   1,  2, 3, 4, 5,   9,     11,    0],
		['1.', '.1', '1.0', '1+1', '+', '-', ''],
	)
	test('int64',
		[None, '42', 0, '-99', '1', 2, 3, 4, 5, '\r9\n', '011', '+0'],
		[None,  42,  0,  -99,   1,  2, 3, 4, 5,   9,       11,    0],
		['1.', '.1', '1.0', '1+1', '+', '-', ''],
	)
	test('bits32',
		['42', 0, '-0', '1', 2, 3, 4, 5, '\r9\n', '011', '+0'],
		[ 42,  0,   0,   1,  2, 3, 4, 5,   9,       11,    0],
		['1.', '.1', '1.0', '1+1', '+', '-', '', '-99'],
		allow_none=False,
	)
	test('bits64',
		['42', 0, '-0', '1', 2, 3, 4, 5, '\r9\n', '011', '+0',],
		[ 42,  0,   0,   1,  2, 3, 4, 5,   9,       11,    0],
		['1.', '.1', '1.0', '1+1', '+', '-', '', '-99'],
		allow_none=False,
	)
	test('json',
		[[1, 2], '[3, 4]', '"foo"', None, 'null', 'NaN', nan, '{"foo": [-Infinity]}'],
		[[1, 2],  [3, 4],   "foo",  None,  None,   nan,  nan,  {'foo': [-inf]}],
		['"foo', '1+1', '', 'None'],
	)



from .printer import prt

description = "Jobs: sliced files, stdout and stderror, load any file from a job, jobwithfile"


def main(urd):
	prt.source(__file__)
	prt()
	prt('This job writes sliced files in "myfile1", and a single file "myfile2".')
	job1 = urd.build('example_writeslicedfile')

	prt()
	prt('These are the files created by / located in job %s.' % (job1,))
	prt.output(job1.files())

	prt()
	prt('This job reads the sliced and the single file and prints its')
	prt('contents to stdout')
	job2 = urd.build(
		'example_readjobwithfile',
		firstfile=job1.withfile('myfile1', sliced=True),
		secondfile=job1.withfile('myfile2'),
	)

	prt()
	prt('Read and print stored stdout from %s synthesis' % (job2,))
	prt.output(job2.output('synthesis'))
	prt()
	prt('Read and print stored stdout from %s everything' % (job2,))
	prt.output(job2.output())
	prt()
	prt('Read and print stored stdout from %s analysis process 2' % (job2,))
	prt.output(job2.output(2))

	prt()
	prt('We can also use job.open() to get a file handle and do any')
	prt('kind of file operations.')
	with job1.open('myfile2', 'rb') as fh:
		import pickle
		prt.output(pickle.load(fh))

	prt('To see all files stored in a job, try')
	prt.command('ax job %s' % (job1,))



from os.path import dirname, join
from accelerator.colour import bold
from .printer import prt

description = "Tutorial: Import a file."


# The file is stored in the same directory as this python file.
# In a real project, set "input directory" in the config file instead!
filename = join(dirname(__file__), 'data.csv')


def main(urd):
	prt.source(__file__)
	prt()
	prt('Import a CSV-file:')
	prt()

	imp = urd.build('csvimport', filename=filename)

	prt()
	with prt.header('HOW TO GET INFORMATION ABOUT JOBS AND DATASETS'):
		prt('To show info on the imported dataset, type')
		prt.command('ax ds', imp)
		prt('To show info on the job that created the dataset, type')
		prt.command('ax job', imp)
		prt('To show the contents of the dataset (with header), type')
		prt.command('ax cat -H', imp)
		prt('Or perhaps in a json format:')
		prt.command('ax cat -f json -D -S -L', imp)

	prt()
	with prt.header('ON DATASET NAMING AND DEFAULT NAME'):
		prt('''
			As long as we do not assign names to datasets, they are
			referenced by the default dataset name, which is "default".
			If we are working on the default dataset, we can use the
			reference to the job that created the dataset as a reference
			to the default dataset inside the job.  I.e.
		''')
		prt.command('ax ds %s' % (imp,))
		prt('is equivalent to')
		prt.command('ax ds %s/default' % (imp,))
		prt('which is the formally correct way to refer to the dataset.')

	prt()
	with prt.header('USING METHOD NAMES INSTEAD OF JOB NAMES.'):
		prt('It is also possible to write for example')
		prt.command('ax job csvimport')
		prt('to get information on the ' + bold('last created') + ' csvimport job.')

	prt()
	with prt.header('GETTING HELP'):
		prt.command('ax <command> --help')
		prt('or just')
		prt.command('ax --help')



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

description = r'''
Create one dataset per value in a column in a dataset (chain).

Reads source back to previous.source.

All datasets in previous get a dataset here, even if empty.
'''

from collections import defaultdict
import itertools
import re

from accelerator.compat import unicode, izip
from accelerator import OptionString, NoSuchDatasetError
from accelerator import subjobs, status


options = {
	'column': OptionString('split based on this column'),
	'length': -1, # Go back at most this many datasets. You almost always want -1 (which goes until previous.source)
}

datasets = ('source',)

jobs = ('previous',)


def prepare(job):
	chain = datasets.source.chain(stop_ds={jobs.previous: 'source'}, length=options.length)
	columns = defaultdict(set)
	none_support = defaultdict(bool)
	hashlabel = set()
	seen_all = set(chain[0].columns)
	for ds in chain:
		if options.column not in ds.columns:
			raise Exception('%r does not have column %r' % (ds, options.column,))
		hashlabel.add(ds.hashlabel)
		seen_all &= set(ds.columns)
		for name, col in ds.columns.items():
			columns[name].add(col.type)
			none_support[name] |= col.none_support
	seen_all.discard(options.column)
	if not seen_all:
		raise Exception('Chain has no common columns (except %r)' % (options.column,))
	columns = {k: columns[k] for k in seen_all}
	if len(hashlabel) == 1:
		hashlabel = hashlabel.pop()
		if hashlabel == options.column:
			hashlabel = None
	else:
		hashlabel = None
	for name, types in columns.items():
		# this relies on internal knowledge that copy_mode is compatible for these types.
		if 'unicode' in types:
			types.discard('ascii')
		if 'number' in types:
			types.discard('bits32')
			types.discard('bits64')
			types.discard('int32')
			types.discard('int64')
			types.discard('float32')
			types.discard('float64')
		if 'bits64' in types:
			types.discard('bits32')
		if 'complex64' in types:
			types.discard('complex32')
		if 'float64' in types:
			types.discard('float32')
		if 'int64' in types:
			types.discard('bits32')
			types.discard('int32')
		if len(types) > 1 and not (types - {'bits32', 'bits64', 'int32', 'int64', 'float32', 'float64'}):
			types = {'number'}
		if len(types) > 1:
			raise Exception("Column %r has incompatible types: %r" % (name, types,))
		columns[name] = (types.pop(), none_support[name],)

	collect = subjobs.build(
		'dataset_fanout_collect',
		source=datasets.source,
		previous=jobs.previous,
		column=options.column,
		length=options.length,
	)
	values = collect.load()
	if jobs.previous:
		previous = {ds.name: ds for ds in jobs.previous.datasets}
		values.update(previous)
	else:
		previous = {}

	with status('Creating %d datasets' % (len(values),)):
		writers = {
			name: job.datasetwriter(
				name=name,
				columns=columns,
				hashlabel=hashlabel,
				previous=previous.get(name),
				copy_mode=True,
			)
			for name in values
		}
	return writers, sorted(columns), chain


def analysis(sliceno, prepare_res):
	writers, columns, chain = prepare_res
	key_it = chain.iterate(sliceno, options.column)
	# we can't just use chain.iterate because of protections against changing types with copy_mode
	values_it = itertools.chain.from_iterable(ds.iterate(sliceno, columns, copy_mode=True, status_reporting=False) for ds in chain)
	for key, values in izip(key_it, values_it):
		writers[unicode(key)].write(*values)



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test copy_mode in DatasetWriter, and three ways to specify the column
types (columns={}, add(n, Datasetcolums), .add(n, (t, none_support)).
'''

from accelerator.dataset import DatasetWriter

jobs = ('source',)


def prepare(job):
	job.datasetwriter(columns=jobs.source.dataset().columns, copy_mode=True)
	DatasetWriter(name='named', columns=jobs.source.dataset('named').columns, copy_mode=True)
	dw_passed = job.datasetwriter(name='passed', copy_mode=True)
	# DatasetColumn in .add
	for n, c in sorted(jobs.source.dataset('passed').columns.items()):
		dw_passed.add(n, c)
	# verify that .add(none_support=) takes precedence over coltype
	dw_nonetest_removed = job.datasetwriter(name='nonetest_removed')
	for n, c in sorted(jobs.source.dataset('nonetest').columns.items()):
		dw_nonetest_removed.add(n, c, none_support=(n == 'unicode'))
	return dw_passed, dw_nonetest_removed

def analysis(sliceno, prepare_res, job):
	dw_default = DatasetWriter()
	dw_named = job.datasetwriter(name='named')
	dw_passed, _ = prepare_res
	for name, dw in [('default', dw_default), ('named', dw_named), ('passed', dw_passed)]:
		for data in jobs.source.dataset(name).iterate(sliceno, copy_mode=True):
			dw.write(*data)

def synthesis(job, slices, prepare_res):
	for name in ['synthesis_split', 'synthesis_manual', 'nonetest']:
		dw = job.datasetwriter(name=name, copy_mode=True)
		# (type, none_support) in .add
		for n, c in sorted(jobs.source.dataset(name).columns.items()):
			dw.add(n, (c.type, c.none_support))
		for sliceno in range(slices):
			dw.set_slice(sliceno)
			for data in jobs.source.dataset(name).iterate(sliceno, copy_mode=True):
				dw.write(*data)
	_, dw_nonetest_removed = prepare_res
	ds = dw_nonetest_removed.finish()
	for name, col in ds.columns.items():
		if name == 'unicode':
			assert col.none_support, "%s:%s should have none_support" % (ds, name,)
		else:
			assert not col.none_support, "%s:%s shouldn't have none_support" % (ds, name,)



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import gzip
import struct
from accelerator.dsutil import typed_writer
from accelerator.subjobs import build

description = r'''
Test the number column type, verifying that both DatasetWriter and
dataset_type produce the expected bitstream.
'''

def synthesis(job, slices):
	# All the representations we want to verify.
	values = [
		# 1 byte values
		[i, '=B', i + 128 + 5]
		for i in range(-5, 123)
	] + [
		# 3 bytes values
		[-6, '=bh', 2, -6],
		[123, '=bh', 2, 123],
		[-0x8000, '=bh', 2, -0x8000],
		[0x7fff, '=bh', 2, 0x7fff],
		# 5 byte values
		[-0x8001, '=bi', 4, -0x8001],
		[0x8000, '=bi', 4, 0x8000],
		[-0x80000000, '=bi', 4, -0x80000000],
		[0x7fffffff, '=bi', 4, 0x7fffffff],
		# 9 byte values
		[-0x80000001, '=bq', 8, -0x80000001],
		[0x80000000, '=bq', 8, 0x80000000],
		[-0x8000000000000000, '=bq', 8, -0x8000000000000000],
		[0x7fffffffffffffff, '=bq', 8, 0x7fffffffffffffff],
		# special values
		[None, '=b', 0],
		[0.1, '=bd', 1, 0.1],
	]

	# Verify each value through a manual typed_writer.
	# Also write to a dataset, a csv and a value2bytes dict.
	value2bytes = {}
	dw = job.datasetwriter()
	dw.add('num', 'number', none_support=True)
	write = dw.get_split_write()
	with job.open('data.csv', 'wt') as csv_fh:
		csv_fh.write('num\n')
		for v in values:
			value = v[0]
			write(value)
			csv_fh.write('%s\n' % (value,))
			want_bytes = struct.pack(*v[1:])
			value2bytes[value] = want_bytes
			with typed_writer('number')('tmp', compression='gzip', none_support=True) as w:
				w.write(value)
			with gzip.open('tmp', 'rb') as fh:
				got_bytes = fh.read()
			assert want_bytes == got_bytes, "%r gave %r, wanted %r" % (value, got_bytes, want_bytes,)

	# Make sure we get the same representation through a dataset.
	# Assumes that the column is merged (a single file for all slices).
	ds = dw.finish()
	just_values = set(v[0] for v in values)
	assert set(ds.iterate(None, 'num')) == just_values, "Dataset contains wrong values"
	want_bytes = b''.join(value2bytes[v] for v in ds.iterate(None, 'num'))
	with gzip.open(ds.column_filename('num'), 'rb') as fh:
		got_bytes = fh.read()
	assert want_bytes == got_bytes, "All individual encoding are right, but not in a dataset?"

	# csvimport and dataset_type the same thing,
	# verify we got the same bytes
	jid = build('csvimport', filename=job.filename('data.csv'))
	jid = build('dataset_type', source=jid, column2type={'num': 'number'}, defaults={'num': None})
	with gzip.open(jid.dataset().column_filename('num'), 'rb') as fh:
		got_bytes = fh.read()
	assert want_bytes == got_bytes, "csvimport + dataset_type (%s) gave different bytes" % (jid,)



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test the range argument to the dataset iteration functions
and chain.with_column().
'''

from datetime import date

date_a = date(2022, 1, 1)
date_b = date(2022, 1, 2)
date_c = date(2022, 1, 3)
date_d = date(2022, 1, 4)

def test(iterfun):
	# verify it contains the expected data
	assert list(iterfun(0, 'a')) == [1, 2, 21, 22, 301, 302]
	assert list(iterfun(0, 'b')) == [1001, 1002, 1021, 1022, 1301, 1302]
	# with the single used column
	assert list(iterfun(0, 'a', range={'a': (2, 22)})) == [2, 21]
	# with the single used column, as a list
	assert list(iterfun(0, ['a'], range={'a': (2, 22)})) == [(2,), (21,)]
	# with the first of several columns
	assert list(iterfun(0, ['a', 'b'], range={'a': (2, 22)})) == [(2, 1002), (21, 1021)]
	# with another of several columns
	assert list(iterfun(0, ['a', 'b'], range={'b': (1002, 1022)})) == [(2, 1002), (21, 1021)]
	# with a column that isn't returned
	assert list(iterfun(0, 'a', range={'b': (1002, 1022)})) == [2, 21]
	# ... as a list
	assert list(iterfun(0, ['a'], range={'b': (1002, 1022)})) == [(2,), (21,)]
	# sloppy range
	assert list(iterfun(0, 'a', range={'a': (2, 22)}, sloppy_range=True)) == [1, 2, 21, 22] # not really guaranteed, but the current implementation does
	# no lower limit
	assert list(iterfun(0, 'a', range={'a': (None, 5)})) == [1, 2]
	# no upper limit
	assert list(iterfun(0, 'a', range={'a': (22, None)})) == [22, 301, 302]
	# no limit at all, despite setting range
	assert list(iterfun(0, 'a', range={'a': (None, None)})) == [1, 2, 21, 22, 301, 302]

def synthesis(job):
	a = job.datasetwriter(name='a', columns={'a': 'int32', 'b': 'int32'}, allow_missing_slices=True)
	b = job.datasetwriter(name='b', columns={'a': 'int32', 'b': 'int64', 'c': 'date'}, allow_missing_slices=True, previous=a)
	c = job.datasetwriter(name='c', columns={'a': 'int64', 'b': 'number', 'c': ('date', True)}, allow_missing_slices=True, previous=b)
	a.set_slice(0)
	a.write(1, 1001)
	a.write(2, 1002)
	a = a.finish()
	b.set_slice(0)
	b.write(21, 1021, date_a)
	b.write(22, 1022, date_b)
	b = b.finish()
	c.set_slice(0)
	c.write(301, 1301, date_c)
	c.write(302, 1302, date_d)
	c = c.finish()

	test(c.iterate_chain)
	# test range on plain ds.iterate
	assert list(c.iterate(0, 'a', range={'a': (None, 302)})) == [301]
	# return several columns but filter on a non-returned one
	assert list(c.iterate(0, ['a', 'c'], range={'b': (None, 1302)})) == [(301, date_c)]
	# filter on a datetime type (a special case in the range check code)
	assert list(c.iterate(0, 'a', range={'c': (None, date_d)})) == [301]
	# really should be the same as c.iterate_chain
	chain = c.chain()
	test(chain.iterate)

	# test the chain range filter
	assert list(chain.range('a', 22, 301).iterate(0, 'a')) == [21, 22]
	assert list(chain.range('a', 22, None).iterate(0, 'a')) == [21, 22, 301, 302]
	assert list(chain.range('a', None, 22).iterate(0, 'a')) == [1, 2, 21, 22]
	# test the chain column filter
	assert list(chain.with_column('c').iterate(0, 'a')) == [21, 22, 301, 302]
	assert list(chain.with_column('b', 'int64').iterate(0, 'a')) == [21, 22]
	assert list(chain.with_column('b', ('int32', 'int64')).iterate(0, 'a')) == [1, 2, 21, 22]
	assert list(chain.with_column('c', none_support=False).iterate(0, 'a')) == [21, 22]
	assert list(chain.with_column('c', none_support=True).iterate(0, 'a')) == [301, 302]



description = "prepare, analysis, and synthesis + return values"


def prepare():
	return 'this is prepare'


def analysis(sliceno, prepare_res):
	return 'this is analysis %s with prepare_res=%s' % (str(sliceno), prepare_res)


def synthesis(analysis_res, prepare_res):
	x = '\n'.join(analysis_res)
	print('analysis_res:', x)
	print('prepare_res:', prepare_res)
	return prepare_res + '   ' + x



from datetime import date, timedelta
from json import dumps
from .printer import prt

description = 'More Urd, several timestamps.'


def main(urd):
	prt.source(__file__)
	# Truncate list to timestamp 0 every time we run this.  If we do
	# not, code changes may lead to inconsistensies that will raise
	# errors.  Per design.  See manual.
	listname = 'example-many'
	urd.truncate(listname, 0)

	prt('Start by creating ten chained jobs.  Each job is associated')
	prt('with a unique Urd-item in the "%s" Urd list.' % (listname,))
	starttime = date(2021, 8, 1)
	for days in range(10):
		timestamp = starttime + timedelta(days=days)
		urd.begin(listname, timestamp, caption='caption_' + str(days))
		previous = urd.latest('listname').joblist.get(-1)
		urd.build('example_dummyjob', timestamp=timestamp, previous=previous)
		urd.finish(listname)

	prt()
	prt('Here\'s a list of all entries in the list "%s"' % (listname,))
	for ts in urd.since(listname, 0):
		prt.output(ts, '"%s"' % (urd.peek(listname, ts).caption,))
	prt('This is available in the shell using')
	prt.command('ax urd %s/' % (listname,))

	prt()
	prt('Here\'s the session for timestamp 2021-08-06')
	prt.plain(dumps(urd.peek(listname, '2021-08-06'), indent=4))
	prt('Note the "deps" section.  The job that is built is fed with a')
	prt('"previous" input job reference.  This reference is provided by Urd')
	prt('using the "urd.latest()" function.')

	prt()
	prt('To print this in a more human friendly format on the command line, try')
	prt.command('    ax urd %s/2021-08-06' % (listname,))



#!/usr/bin/env python

############################################################################
#                                                                          #
# Copyright (c) 2019-2021 Carl Drougge                                     #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function

from setuptools import setup, find_packages, Extension
from importlib import import_module
from os.path import exists
import os
from datetime import datetime
from subprocess import check_output, check_call, CalledProcessError
from io import open
import re
import sys

if os.name != 'posix':
	print("Accelerator only works on posix platforms (like Linux).", file=sys.stderr)
	if os.name == 'nt':
		print("You could try running under WSL.", file=sys.stderr)
	sys.exit(1)

def dirty():
	for extra in ([], ['--cached'],):
		cmd = ['git', 'diff-index', '--quiet', 'HEAD'] + extra
		try:
			check_call(cmd)
		except CalledProcessError as e:
			if e.returncode == 1:
				return '.dirty'
			else:
				raise
	return ''

with open('README.md', 'r', encoding='utf-8') as fh:
	long_description = []
	for line in fh:
		if not line.startswith('[PyPI]') and not line.startswith('`pip'):
			long_description.append(line)
	long_description = ''.join(long_description)

if exists('PKG-INFO'):
	with open('PKG-INFO', 'r', encoding='utf-8') as fh:
		for line in fh:
			if line.startswith('Version: '):
				version = line.strip().split()[1]
				break
else:
	version = datetime.utcnow().strftime('%Y.%m.%d')
	env_version = os.environ.get('ACCELERATOR_BUILD_VERSION')
	if env_version and dirty():
		raise Exception("Refusing to build a specified version from a dirty repo")
	commit = check_output(['git', 'rev-parse', 'HEAD']).strip().decode('ascii')
	if os.environ.get('ACCELERATOR_BUILD') == 'IS_RELEASE':
		if not env_version:
			raise Exception("Refusing to build release without a specified version")
		assert re.match(r'20\d\d\.\d\d\.\d\d$', env_version)
		version = env_version
	else:
		if env_version:
			assert re.match(r'20\d\d\.\d\d\.\d\d\.dev\d+$', env_version)
			version = env_version
		else:
			version = "%s.dev1+%s%s" % (version, commit[:10], dirty(),)
	version = version.replace('.0', '.')
	with open('accelerator/version.txt', 'w') as fh:
		fh.write(version + '\n')
		fh.write(commit + dirty() + '\n')

def mk_file(fn, contents):
	if exists(fn):
		with open(fn, 'rb') as fh:
			old_contents = fh.read().decode('utf-8')
	else:
		old_contents = None
	if contents != old_contents:
		with open(fn, 'wb') as fh:
			fh.write(contents.encode('utf-8'))
	return fn

def mk_ext(name, *sources):
	zlib = os.environ.get('ACCELERATOR_BUILD_STATIC_ZLIB')
	if zlib:
		kw = dict(extra_objects=[zlib])
	else:
		kw = dict(libraries=['z'])
	return Extension(
		name,
		sources=list(sources),
		extra_compile_args=['-std=c99', '-O3', '-fvisibility=hidden'],
		**kw
	)

dsutilmodule = mk_ext('accelerator._dsutil', 'dsutil/siphash24.c', 'dsutil/dsutilmodule.c')

def method_mod(name):
	code = import_module('accelerator.standard_methods.' + name).c_module_code
	fn = 'accelerator/standard_methods/_generated_' + name + '.c'
	return mk_ext('accelerator.standard_methods._' + name, mk_file(fn, code))

dataset_typemodule = method_mod('dataset_type')
csvimportmodule = method_mod('csvimport')

setup(
	name="accelerator",
	version=version,
	packages=find_packages(),

	entry_points={
		'console_scripts': [
			'ax = accelerator.shell:main',
		],
	},

	install_requires=[
		'setproctitle>=1.1.8', # not actually required
		'bottle>=0.12.7, <0.13',
		'waitress>=1.0',
		'configparser>=3.5.0, <5.0; python_version<"3"',
		'monotonic>=1.0; python_version<"3"',
		'selectors2>=2.0; python_version<"3"',
	],
	python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, <4",

	ext_modules=[dsutilmodule, dataset_typemodule, csvimportmodule],

	package_data={
		'': ['*.txt', 'methods.conf', 'board/*.tpl', 'examples/*.csv'],
	},

	author="Carl Drougge",
	author_email="bearded@longhaired.org",
	url="https://exax.org/",
	description="A tool for fast and reproducible processing of large amounts of data.",
	long_description=long_description,
	long_description_content_type="text/markdown",
	project_urls={
		"Source": "https://github.com/eBay/accelerator",
		"Reference manual": "https://berkeman.github.io/pdf/acc_manual.pdf",
	},

	classifiers=[
		"Development Status :: 4 - Beta",
		"Environment :: Console",
		"Intended Audience :: Developers",
		"Intended Audience :: Science/Research",
		"License :: OSI Approved :: Apache Software License",
		"Operating System :: POSIX",
		"Operating System :: POSIX :: BSD :: FreeBSD",
		"Operating System :: POSIX :: Linux",
		"Programming Language :: Python :: 2",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: Implementation :: CPython",
		"Programming Language :: C",
	],
)



############################################################################
#                                                                          #
# Copyright (c) 2021-2022 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test the alias and colour config for shell commands.
'''

options = dict(
	command_prefix=['ax', '--config', '/some/path/here'],
)

import os
from subprocess import check_output, CalledProcessError

config = '''
[alias]
	path = job -P

[colour]
'''

def ax_cmd(*a):
	cmd = options.command_prefix + list(a)
	res = check_output(cmd)
	res = res.decode('utf-8', 'replace')
	return res.strip()

def synthesis(job):
	os.putenv('XDG_CONFIG_HOME', job.path)
	os.putenv('CLICOLOR_FORCE', '1')
	os.mkdir('accelerator')
	with open('accelerator/config', 'w') as fh:
		fh.write(config)
	# test an alias
	assert ax_cmd('path', job) == job.path

	def chk(cfgstr, pre, post):
		with open('accelerator/config', 'w') as fh:
			fh.write(config)
			fh.write(cfgstr)
		try:
			got = ax_cmd('job', job).split('\n')[-1]
		except CalledProcessError as e:
			raise Exception("%r could not run: %s" % (cfgstr, e,))
		want = '%sWARNING: Job did not finish%s' % (pre, post,)
		if got != want:
			raise Exception("%r:\nWanted %r\ngot    %r" % (cfgstr, want, got,))

	# test the colour config
	chk('\twarning = CYAN', '\x1b[36m', '\x1b[39m')
	# test that a more specific colour config wins
	chk('\twarning = CYAN\n\tjob/warning = BOLD #030405', '\x1b[1;38:2:3:4:5m', '\x1b[22;39m')
	# test literals in the config
	chk('\twarning = <FOO >BAR', 'FOO', 'BAR')
	# combining literals with named attributes, and escapes
	chk(
		'\twarning = <FOO >BAR\n\tjob/warning = BOLD BLUE <\\x18\\\\\\eFOO\T >\\?\\\\eBAR\E UNDERLINE',
		'\x1b[1;34m\x18\\\x1bFOO\t\x1b[4m',
		'\x1b[22;24;39m\\?\\eBAR\x1b',
	)

	# test various other colour formats in config
	for cfgstr, pre, post in (
		('\twarning = #123', '\x1b[38:5:67m', '\x1b[39m'),
		('\twarning = #0 #1cbg', '\x1b[38:5:16;48:5:255m', '\x1b[39;49m'),
		('\twarning = #1d Xffbg', '\x1b[38:5:231;48:5:255m', '\x1b[39;49m'),
		('\twarning = X10 #20ff00bg', '\x1b[38:5:16;48:2:32:255:0m', '\x1b[39;49m'),
	):
		chk(cfgstr, pre, post)



############################################################################
#                                                                          #
# Copyright (c) 2021 Anders Berkeman                                       #
# Modifications copyright (c) 2021 Carl Drougge                            #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import sys
from glob import glob
from os.path import dirname, basename
from importlib import import_module
from accelerator.compat import terminal_size, ArgumentParser
from accelerator.colourwrapper import colour
from accelerator.shell import printdesc


def main(argv, cfg):
	descr = "lists and describes build scripts"
	parser = ArgumentParser(
		prog=argv.pop(0),
		description=descr,
	)
	parser.add_argument('-s', '--short', action='store_true', help='short listing')
	parser.add_argument('-p', '--path', action='store_true', help='show package paths')
	parser.add_argument('match', nargs='*', default=[], help='substring used for matching')
	args = parser.parse_intermixed_args(argv)
	columns = terminal_size().columns

	if not args.match:
		# no args => list everything in short format
		args.match = ['']
		args.short = True

	packages = []
	for package in cfg.method_directories:
		path = dirname(import_module(package).__file__)
		scripts = []
		packages.append((package, path, scripts))
		for item in sorted(glob(path + '/build.py') + glob(path + '/build_*.py')):
			name = basename(item[:-3])
			modname = '.'.join((package, name))
			if any(m in modname for m in args.match):
				try:
					module = import_module(modname)
				except Exception as e:
					print(colour('%s: %s' % (item, e,), 'script/warning'), file=sys.stderr)
					continue
				scripts.append((name, getattr(module, 'description', '')))

	for package, path, scripts in sorted(packages):
		if scripts:
			if args.path:
				print(path + '/')
			else:
				print(package)
			printdesc(sorted(scripts), columns, 'script', full=not args.short)



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2020 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

description = r'''
Take a chain of datasets and make a checksum of one or more columns.
See dataset_checksum.description for more information.

datasets.source is mandatory, datasets.stop is optional.
options.chain_length defaults to -1.

Sort does not sort across datasets.
'''

from accelerator import DotDict, build

options = dict(
	chain_length = -1,
	columns      = set(),
	sort         = True,
)

datasets = ('source', 'stop',)

def synthesis():
	sum = 0
	jobs = datasets.source.chain(length=options.chain_length, stop_ds=datasets.stop)
	for src in jobs:
		data = build('dataset_checksum', columns=options.columns, sort=options.sort, source=src).load()
		sum ^= data.sum
	print("Total: %016x" % (sum,))
	return DotDict(sum=sum, columns=data.columns, sort=options.sort, sources=jobs)



from .printer import prt

description = 'Basic Urd database useage'


def main(urd):
	prt.source(__file__)
	# Truncate list to timestamp 0 every time we run this.  If we do
	# not, code changes may lead to inconsistensies that will raise
	# errors.  Per design.  See manual.
	listname = 'example-basic'
	timestamp = '2021-07-01'
	urd.truncate(listname, 0)

	prt()
	prt('Create and Urd item containing a joblist of one job.')
	prt('The name of the Urd list is "%s".' % (listname,))
	urd.begin(listname, timestamp, caption='you may assign a caption here')
	urd.build('example_returnhelloworld')
	urd.finish(listname)

	prt()
	prt('Here\'s a list of all available Urd-lists')
	for item in urd.list():
		prt.output(item)

	prt()
	prt('The command')
	prt.command('ax urd')
	prt('# will do the same thing on the command line.')

	prt()
	prt('Here\'s a list of all entries in the list "%s"' % (listname,))
	for ts in urd.since(listname, 0):
		prt.output(ts)

	prt()
	prt('The command')
	prt.command('ax urd %s/since/0' % (listname,))
	prt('or')
	prt.command('ax urd %s/' % (listname,))
	prt('will do the same thing on the command line.')

	prt()
	prt('To see a specific entry, try')
	prt.command('ax urd %s/%s' % (listname, timestamp))

	prt()
	prt('To see information about the a specific job in that Urd session')
	prt.command('ax job :%s/%s:example1' % (listname, timestamp))

	prt()
	prt('To see information about the last job in that Urd session')
	prt.command('ax job :%s/%s:-1' % (listname, timestamp))
	prt.command('ax job :%s/%s:' % (listname, timestamp))
	prt('Note that the commands defaults to the _last_ item')

	prt()
	prt('To see information about the last job in the last Urd session')
	prt.command('ax job :%s:' % (listname,))
	prt('Note that the commands defaults to the _last_ item')

	prt()
	prt('This is particularly useful for dataset, where we can write for')
	prt('example "ax cat :import:"  to print the dataset in the last job')
	prt('in the last Urd-item in the "import" list.')



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test that missing set_slice is an error without allow_missing_slices
but not with.
'''

from accelerator.error import DatasetUsageError

def synthesis(job):
	dw = job.datasetwriter(name='fails')
	dw.add('a', 'ascii')
	dw.set_slice(1)
	dw.write('a')
	try:
		dw.finish()
		raise Exception("DatasetWriter allowed finishing without writing all slices")
	except DatasetUsageError:
		dw.discard()

	dw = job.datasetwriter(name='fails', allow_missing_slices=True)
	dw.add('a', 'ascii')
	try:
		dw.get_split_write()
		raise Exception("DatasetWriter allowed get_split_write with allow_missing_slices")
	except DatasetUsageError:
		dw.discard()

	dw = job.datasetwriter(name='succeeds', allow_missing_slices=True)
	dw.add('a', 'ascii')
	dw.set_slice(1)
	dw.write('a')
	dw.finish()



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2022 Carl Drougge                       #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division

from glob import glob
from collections import defaultdict
from bottle import route, request, auth_basic, abort
import bottle
from threading import Lock
import json
import re
from datetime import datetime
import operator
import os.path
from io import TextIOWrapper
import sys
import os
import signal

from accelerator.compat import iteritems, itervalues, unicode, ArgumentParser
from accelerator.colourwrapper import colour
from accelerator.extras import DotDict, PY3
from accelerator.unixhttp import WaitressServer

LOGFILEVERSION = '3'

lock = Lock()

def locked(func):
	def inner(*a, **kw):
		with lock:
			return func(*a, **kw)
	return inner


def joblistlike(jl):
	assert isinstance(jl, list)
	for v in jl:
		assert isinstance(v, (list, tuple)), v
		assert len(v) == 2, v
		for s in v:
			assert isinstance(s, unicode), s
	return True


class TimeStamp(str):
	"""Can be a string like 2019-12-09T12:19:04.123456 (day and time are
	optional, partial time ok) and/or an int >= 0.
	Integers without datetimes sort before all datetimes.
	Datetimes without integers sort before the same datetime with an integer.
	When both are specified they are separated by a +
	"""

	if PY3: # python2 doesn't support slots on str subclasses
		__slots__ = ('_ts', '_integer')

	def __new__(cls, ts):
		if isinstance(ts, TimeStamp):
			return ts
		try:
			integer = int(ts, 10)
			assert integer >= 0, 'Invalid timestamp %d' % (ts,)
			ts = None
		except ValueError:
			m = re.match(r'(\d{4}-\d{2}(?:-\d{2}(?:[T ]\d{2}(?::\d{2}(?::\d{2}(?:\.\d{1,6})?)?)?)?)?)(\+\d+)?$', ts)
			assert m, 'Invalid timestamp %s' % (ts,)
			ts, integer = m.groups()
			ts = ts.replace(' ', 'T')
			integer = int(integer[1:], 10) if integer else None
		assert ts is not None or integer is not None, 'Invalid timestamp %s' % (ts,)
		if ts:
			if integer is not None:
				strval = '%s+%d' % (ts, integer,)
			else:
				strval = ts
		else:
			strval = str(integer)
		obj = str.__new__(cls, strval)
		obj._ts = ts
		obj._integer = integer
		return obj

	__hash__ = str.__hash__

	def __eq__(self, other):
		if not isinstance(other, TimeStamp):
			other = TimeStamp(other)
		return self._ts == other._ts and self._integer == other._integer

	def __lt__(self, other):
		if not isinstance(other, TimeStamp):
			other = TimeStamp(other)
		if self._ts is not None:
			if other._ts is not None:
				if self._integer is not None:
					if self._ts == other._ts:
						if other._integer is not None:
							return self._integer < other._integer
						else:
							return False
					else:
						return self._ts < other._ts
				else:
					if other._integer is not None:
						return self._ts <= other._ts
					else:
						return self._ts < other._ts
			else:
				return False
		elif other._ts is not None:
			return True
		else:
			return self._integer < other._integer

	def __le__(self, other):
		if not isinstance(other, TimeStamp):
			other = TimeStamp(other)
		return self < other or self == other

	def __ge__(self, other):
		return not self < other

	def __gt__(self, other):
		return not self <= other


class DB:
	def __init__(self, path, verbose=True):
		self._initialised = False
		self.path = path
		self.db = defaultdict(dict)
		self.ghost_db = defaultdict(lambda: defaultdict(list))
		if os.path.isdir(path):
			files = glob(os.path.join(path, '*/*.urd'))
			self._parsed = {}
			stat = {}
			for fn in files:
				with open(fn) as fh:
					ix = 0
					for line in fh:
						self._parse(line)
						ix += 1
					stat[fn[len(path) + 1:-len('.urd')]] = ix
			self._playback_parsed()
			if verbose:
				print("urd-list                          lines     ghosts     active")
				for key, val in sorted(stat.items()):
					print("%-30s  %7d    %7d    %7d" % (key, val, len(self.ghost_db[key]), len(self.db[key]),))
				print()
		else:
			print("Creating directory \"%s\"." % (path,))
			os.makedirs(path)
		self._lasttime = None
		self._initialised = True

	def _parse(self, line):
		line = line.rstrip('\n').split('|')
		logfileversion, writets = line[:2]
		assert logfileversion == '3'
		assert writets not in self._parsed
		self._parsed[writets] = line[2:]

	def _playback_parsed(self):
		for _writets, line in sorted(iteritems(self._parsed)):
			action = line.pop(0)
			assert action in ('add', 'truncate',)
			if action == 'add':
				self._parse_add(line)
			elif action == 'truncate':
				self._parse_truncate(line)
			else:
				assert "can't happen"

	def _parse_add(self, line):
		key = line[1]
		user, build = key.split('/')
		flags = line[4].split(',') if line[4] else []
		data = DotDict(timestamp=line[0],
			user=user,
			build=build,
			deps=json.loads(line[2]),
			joblist=json.loads(line[3]),
			flags=flags,
			caption=line[5],
		)
		self.add(data)

	def _parse_truncate(self, line):
		timestamp, key = line
		self.truncate(key, timestamp)

	def _validate_data(self, data, with_deps=True):
		if with_deps:
			assert set(data) == {'timestamp', 'joblist', 'caption', 'user', 'build', 'deps', 'flags',}
			assert isinstance(data.user, unicode)
			assert isinstance(data.build, unicode)
			assert isinstance(data.deps, dict)
			for v in itervalues(data.deps):
				assert isinstance(v, dict)
				self._validate_data(DotDict(v), False)
		else:
			assert set(data) == {'timestamp', 'joblist', 'caption',}
		assert joblistlike(data.joblist), data.joblist
		assert data.joblist
		assert isinstance(data.caption, unicode)
		data.timestamp = TimeStamp(data.timestamp)

	def _serialise(self, action, data):
		if action == 'add':
			self._validate_data(data)
			json_deps = json.dumps(data.deps)
			json_joblist = json.dumps(data.joblist)
			key = '%s/%s' % (data.user, data.build,)
			flags = ','.join(data.flags)
			for s in json_deps, json_joblist, data.caption, data.user, data.build, data.timestamp, flags:
				assert '|' not in s, s
			logdata = [json_deps, json_joblist, flags, data.caption,]
		elif action == 'truncate':
			key = data.key
			logdata = []
		else:
			assert "can't happen"
		data.timestamp = TimeStamp(data.timestamp)
		while True: # paranoia
			now = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")
			if now != self._lasttime: break
		self._lasttime = now
		s = '|'.join([LOGFILEVERSION, now, action, data.timestamp, key,] + logdata)
		return s

	def _is_ghost(self, data):
		for key, data in iteritems(data.deps):
			db = self.db[key]
			ts = data['timestamp']
			if ts not in db:
				return True
			for k, v in iteritems(data):
				if db[ts].get(k) != v:
					return True
		return False

	@locked
	def add(self, data):
		key = '%s/%s' % (data.user, data.build)
		flags = data.pop('flags', [])
		assert flags in ([], ['update']), 'Unknown flags: %r' % (flags,)
		new = False
		changed = False
		ghosted = 0
		data.timestamp = TimeStamp(data.timestamp)
		assert data.timestamp != '0', "Timestamp 0 is special, you can't add it."
		is_ghost = self._is_ghost(data)
		if is_ghost:
			db = self.ghost_db[key]
			if data.timestamp in db:
				new = False
				changed = (data not in db[data.timestamp])
			else:
				new = True
		else:
			db = self.db[key]
			if data.timestamp in db:
				new = False
				changed = (db[data.timestamp] != data)
			else:
				new = True
		if changed and 'update' not in flags:
			assert self._initialised, 'Log updates without update flag: %r' % (data,)
			bottle.response.status = 409
			return {'error': 'would update'}
		if new or changed:
			data.flags = flags
			self.log('add', data) # validates, too
			del data.flags
			if is_ghost:
				db[data.timestamp].append(data)
			else:
				if changed:
					ghost_data = db[data.timestamp]
					self.ghost_db[key][data.timestamp].append(ghost_data)
				db[data.timestamp] = data
				if changed:
					ghosted = self._update_ghosts()
		res = dict(new=new, changed=changed, is_ghost=is_ghost)
		if changed:
			res['deps'] = ghosted
		return res

	def _update_ghosts(self):
		def inner():
			count = 0
			for key, db in iteritems(self.db):
				for ts, data in sorted(db.items()):
					if self._is_ghost(data):
						count += 1
						del db[ts]
						self.ghost_db[key][ts].append(data)
			return count
		res = 0
		while True:
			count = inner()
			if not count:
				break
			res += count
		return res

	@locked
	def truncate(self, key, timestamp):
		old = self.db[key]
		new = {}
		ghost = {}
		timestamp = TimeStamp(timestamp)
		for ts, data in iteritems(old):
			if ts < timestamp:
				new[ts] = data
			else:
				ghost[ts] = data
		self.log('truncate', DotDict(key=key, timestamp=timestamp))
		self.db[key] = new
		ghost_db = self.ghost_db[key]
		for ts, data in iteritems(ghost):
			ghost_db[ts].append(data)
		if ghost:
			deps = self._update_ghosts()
		else:
			deps = 0
		return {'count': len(ghost), 'deps': deps}

	def log(self, action, data):
		if self._initialised:
			if action == 'truncate':
				user, build = data.key.split('/')
			else:
				user, build = data.user, data.build
			assert '/' not in user
			assert '/' not in build
			path = os.path.join(self.path, user)
			if not os.path.isdir(path):
				os.makedirs(path)
			fn = os.path.join(path, build + '.urd')
			with open(fn, 'a') as fh:
				start_pos = fh.tell()
				try:
					fh.write(self._serialise(action, data) + '\n')
					fh.flush()
				except IOError as e:
					try:
						try:
							# Try to avoid leaving a partial line in the file.
							fh.truncate(start_pos)
							fh.close()
							extra = ''
						except:
							extra = "  Also failed to remove partially written data."
							stars = colour.red('****')
							extra2 = "  " + stars + " YOUR URD DB IS PROBABLY BROKEN NOW! " + stars
						msg = "  Failed to write %s: %s" % (fn, e)
						brk = "#" * (max(len(msg), len(extra)) + 2)
						print("", file=sys.stderr)
						print(brk, file=sys.stderr)
						print(msg, file=sys.stderr)
						if extra:
							print(extra, file=sys.stderr)
							print(extra2, file=sys.stderr)
						print(brk, file=sys.stderr)
						print("", file=sys.stderr)
					finally:
						# This is a fatal error.
						os.killpg(os.getpgid(0), signal.SIGTERM)

	@locked
	def get(self, key, timestamp):
		db = self.db[key]
		return db.get(TimeStamp(timestamp))

	@locked
	def since(self, key, timestamp):
		timestamp = TimeStamp(timestamp)
		return sorted(k for k in self.db[key] if k > timestamp)

	@locked
	def limited_endpoint(self, key, timestamp, cmpfunc, minmaxfunc):
		db = self.db[key]
		try:
			k = minmaxfunc(k for k in db if cmpfunc(k, timestamp))
		except ValueError: # empty sequence
			return None
		return db[k]

	@locked
	def latest(self, key):
		db = self.db[key]
		if db:
			return db[max(db)]

	@locked
	def first(self, key):
		db = self.db[key]
		if db:
			return db[min(db)]

	def keys(self):
		return filter(self.db.get, self.db)


def auth(user, passphrase):
	return authdict.get(user) == passphrase or (allow_passwordless and user not in authdict)

def timestamp404(timestamp):
	try:
		return TimeStamp(timestamp)
	except AssertionError as e:
		raise bottle.HTTPError(404, str(e))

@route('/<user>/<build>/since/<timestamp>')
def since(user, build, timestamp):
	key = user + '/' + build
	timestamp = timestamp404(timestamp)
	res = db.since(key, timestamp)
	if 'captions' in request.query:
		res = [(ts, (db.get(key, ts) or {}).get('caption', ''),) for ts in res]
	return res

@route('/<user>/<build>/latest')
def latest(user, build):
	return db.latest(user + '/' + build)

@route('/<user>/<build>/first')
def first(user, build):
	return db.first(user + '/' + build)

@route('/<user>/<build>/<timestamp>')
def single(user, build, timestamp):
	key = user + '/' + build
	if len(timestamp) > 1 and timestamp[0] in '<>':
		if timestamp[0] == '<':
			minmaxfunc = max
			if timestamp[1] == '=':
				if '-' in timestamp:
					# we want 2014-04-10 <= 2014-04 to be True
					def cmpfunc(k, ts):
						return k <= ts or k.startswith(ts)
				else:
					cmpfunc = operator.le
				timestamp = timestamp[2:]
			else:
				cmpfunc = operator.lt
				timestamp = timestamp[1:]
		else:
			minmaxfunc = min
			if timestamp[1] == '=':
				cmpfunc = operator.ge
				timestamp = timestamp[2:]
			else:
				cmpfunc = operator.gt
				timestamp = timestamp[1:]
		timestamp = timestamp404(timestamp)
		return db.limited_endpoint(key, timestamp, cmpfunc, minmaxfunc)
	else:
		timestamp = timestamp404(timestamp)
		return db.get(key, timestamp)


@route('/add', method='POST')
@auth_basic(auth)
def add():
	body = request.body
	if PY3:
		body = TextIOWrapper(body, encoding='utf-8')
	data = DotDict(json.load(body))
	if data.user != request.auth[0]:
		abort(401, "Error:  user does not match authentication!")
	result = db.add(data)
	return result


@route('/truncate/<user>/<build>/<timestamp>', method='POST')
@auth_basic(auth)
def truncate(user, build, timestamp):
	if user != request.auth[0]:
		abort(401, "Error:  user does not match authentication!")
	timestamp = timestamp404(timestamp)
	return db.truncate(user + '/' + build, timestamp)


@route('/test/<user>', method='POST')
@auth_basic(auth)
def test(user):
	if user != request.auth[0]:
		abort(401, "Error:  user does not match authentication!")


@route('/list')
def slash_list():
	return sorted(db.keys())


@bottle.error(401)
@bottle.error(404)
@bottle.error(409)
@bottle.error(500)
def error_handler(e):
	res = [e.body]
	if e.exception:
		res.append(repr(e.exception))
	if e.traceback:
		res.append(e.traceback)
	if isinstance(e.exception, AssertionError):
		# Just the message then
		res = [str(e.exception)]
	bottle.response.content_type = 'text/plain'
	return '\n'.join(res).encode('utf-8')


def readauth(filename):
	if not os.path.exists(filename):
		return {}
	d = {}
	with open(filename) as fh:
		for line in fh:
			line = line.strip()
			if not line or line.startswith('#'):  continue
			line = line.split(':')
			assert len(line) == 2, "Parse error in \"" + filename + "\" " +  ':'.join(line)
			d[line[0]] = line[1]
	return d


def jsonify(callback):
	def func(*a, **kw):
		res = callback(*a, **kw)
		if isinstance(res, (bottle.BaseResponse, bottle.BottleException)):
			return res
		bottle.response.content_type = 'text/json'
		return json.dumps(res)
	return func


def main(argv, cfg):
	global authdict, allow_passwordless, db

	parser = ArgumentParser(prog=argv.pop(0))
	parser.add_argument('--path', type=str, default='urd.db',
		help='database directory (can be relative to project directory) (default: urd.db)',
	)
	parser.add_argument('--allow-passwordless', action='store_true', help='accept any pass for users not in passwd.')
	parser.add_argument('--quiet', action='store_true', help='less chatty.')
	args = parser.parse_args(argv)
	if not args.quiet:
		print('-'*79)
		print(args)
		print()

	auth_fn = os.path.join(args.path, 'passwd')
	authdict = readauth(auth_fn)
	allow_passwordless = args.allow_passwordless
	if not authdict and not args.allow_passwordless:
		raise Exception('No users in %r and --allow-passwordless not specified.' % (auth_fn,))
	db = DB(args.path, not args.quiet)

	bottle.install(jsonify)

	kw = dict(debug=False, reloader=False, quiet=args.quiet, server=WaitressServer)
	listen = cfg.urd_listen
	if not listen:
		raise Exception('urd not configured in this project')
	if isinstance(listen, tuple):
		kw['host'], kw['port'] = listen
	else:
		from accelerator.server import check_socket
		check_socket(listen)
		kw['host'] = listen
		kw['port'] = 0
	bottle.run(**kw)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2019-2022 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify various corner cases in dataset_type.
'''

from datetime import date, time, datetime
from math import isnan
import json
import sys

from accelerator import subjobs
from accelerator.dispatch import JobError
from accelerator.dataset import Dataset, DatasetWriter
from accelerator.dsutil import typed_writer
from accelerator.compat import PY3
from accelerator.standard_methods import dataset_type
from accelerator import g

options = {'numeric_comma': True}

depend_extra = (dataset_type,)

all_typenames = set(dataset_type.convfuncs)
used_typenames = set()

def used_type(typ):
	if ':' in typ and typ not in all_typenames:
		typ = typ.split(':', 1)[0] + ':*'
	used_typenames.add(typ)

no_default = object()
def verify(name, types, bytes_data, want, default=no_default, want_fail=False, all_source_types=False, **kw):
	todo = [('bytes', bytes_data,)]
	if all_source_types:
		uni_data = [v.decode('ascii') for v in bytes_data]
		todo += [('ascii', uni_data,), ('unicode', uni_data,)]
	for coltype, data in todo:
		dsname = '%s %s' % (name, coltype,)
		_verify(dsname, types, data, coltype, want, default, want_fail, kw)

def _verify(name, types, data, coltype, want, default, want_fail, kw):
	if callable(want):
		check = want
	else:
		def check(got, fromstr, filtered=False):
			want1 = want if isinstance(want, list) else want[typ]
			if filtered:
				want1 = want1[::2]
			assert got == want1, 'Expected %r, got %r from %s.' % (want1, got, fromstr,)
	dw = DatasetWriter(name=name, columns={'data': coltype, 'extra': 'bytes'})
	dw.set_slice(0)
	for ix, v in enumerate(data):
		dw.write(v, b'1' if ix % 2 == 0 else b'skip')
	for sliceno in range(1, g.slices):
		dw.set_slice(sliceno)
	bytes_ds = dw.finish()
	for typ in types:
		opts = dict(column2type=dict(data=typ))
		opts.update(kw)
		if default is not no_default:
			opts['defaults'] = {'data': default}
		try:
			jid = subjobs.build('dataset_type', datasets=dict(source=bytes_ds), options=opts)
		except JobError:
			if want_fail:
				continue
			raise Exception('Typing %r as %s failed.' % (bytes_ds, typ,))
		assert not want_fail, "Typing %r as %s should have failed, but didn't (%s)." % (bytes_ds, typ, jid)
		typed_ds = Dataset(jid)
		got = list(typed_ds.iterate(0, 'data'))
		check(got, '%s (typed as %s from %r)' % (typed_ds, typ, bytes_ds,))
		if opts.get('filter_bad'):
			bad_ds = Dataset(jid, 'bad')
			got_bad = list(bad_ds.iterate(0, 'data'))
			assert got_bad == [b'nah'], "%s should have had a single b'nah', but had %r" % (bad_ds, got_bad,)
		if 'filter_bad' not in opts and not callable(want):
			opts['filter_bad'] = True
			opts['column2type']['extra'] = 'int32_10'
			jid = subjobs.build('dataset_type', datasets=dict(source=bytes_ds), options=opts)
			typed_ds = Dataset(jid)
			got = list(typed_ds.iterate(0, 'data'))
			check(got, '%s (typed as %s from %r with every other line skipped from filter_bad)' % (typed_ds, typ, bytes_ds,), True)
			want_bad = [t for t in bytes_ds.iterate(0) if t[1] == b'skip']
			bad_ds = Dataset(jid, 'bad')
			got_bad = list(bad_ds.iterate(0))
			assert got_bad == want_bad, "Expected %r, got %r from %s" % (want_bad, got_bad, bad_ds,)
		used_type(typ)

def test_numbers():
	verify('floats', ['complex32', 'complex64', 'float32', 'float64', 'number'], [b'1.50', b'-inf', b'5e-1'], [1.5, float('-inf'), 0.5], all_source_types=True)
	if options.numeric_comma:
		verify('numeric_comma', ['complex32', 'complex64', 'float32', 'float64', 'number'], [b'1,5', b'1.0', b'9'], [1.5, 42.0, 9.0], '42', numeric_comma=True)
	verify('float32 rounds', ['complex32', 'float32'], [b'1.2'], [1.2000000476837158])
	verify('filter_bad', ['int32_10', 'int64_10', 'bits32_10', 'bits64_10', 'complex32', 'complex64', 'float32', 'float64', 'number'], [b'4', b'nah', b'1', b'0'], [4, 1, 0], filter_bad=True)

	all_source_types = True
	for base, values in (
		(10, (b'27', b'027', b' \r27 '),),
		( 8, (b'33', b'033', b'\t000033'),),
		(16, (b'1b', b'0x1b', b'\r001b',),),
		( 0, (b'27', b'\r033', b'0x1b',),),
	):
		types = ['%s_%d' % (typ, base,) for typ in ('int32', 'int64', 'bits32', 'bits64')]
		verify('base %d' % (base,), types, values, [27, 27, 27], all_source_types=all_source_types)
		types = [typ + 'i' for typ in types]
		if base == 10:
			types += ['float32i', 'float64i']
		values = [v + b'garbage' for v in values]
		verify('base %d i' % (base,), types, values, [27, 27, 27], all_source_types=all_source_types)
		all_source_types = False
	verify('inty numbers', ['number', 'number:int'], [b'42', b'42.0', b'42.0000000', b'43.'], [42, 42, 42, 43])
	if options.numeric_comma:
		verify('inty numbers numeric_comma', ['number', 'number:int'], [b'42', b'42,0', b'42,0000000', b'43,'], [42, 42, 42, 43], numeric_comma=True)

	# Python 2 accepts 42L as an integer, python 3 doesn't. The number
	# type falls back to python parsing, verify this works properly.
	verify('integer with L', ['number'], [b'42L'], [42], want_fail=PY3)

	# tests both that values outside the range are rejected
	# and that None works as a default value for relevant types.
	for typ, values, default in (
		('int32_10', (b'2147483648', b'-2147483648', b'1O',), '123',),
		('int32_16', (b'80000000', b'-80000000', b'1O',), None,),
		('int64_10', (b'36893488147419103231', b'9223372036854775808', b'-9223372036854775808', b'1O',), None,),
		('bits32_10', (b'4294967296', b'8589934591', b'-1', b'1O',), '456',),
		('bits64_10', (b'36893488147419103231', b'18446744073709551616', b'-1', b'1O',), '789',),
	):
		if default is None:
			want = [None] * len(values)
		else:
			want = [int(default)] * len(values)
		verify('nearly good numbers ' + typ, [typ], values, want, default)

	verify('not a number', ['number'], [b'forty two'], [42], want_fail=True)

	verify('strbool', ['strbool'], [b'', b'0', b'FALSE', b'f', b'FaLSe', b'no', b'off', b'NIL', b'NULL', b'y', b'jao', b'well, sure', b' ', b'true'], [False] * 9 + [True] * 5)
	verify('floatbool false', ['floatbool'], [b'0', b'-0', b'1', b'1004', b'0.00001', b'inf', b'-1', b'', b'0.00'], [False, False, True, True, True, True, True, False, False])
	verify('floatbool i', ['floatbooli'], [b'1 yes', b'0 no', b'0.00 also no', b'inf yes', b' 0.01y'], [True, False, False, True, True])
	for typ, smallbig, smallneg, bigbig, bigneg, extra in (
		('floatint32e', 42, 42, 42, 42, 42,),
		('floatint32ei', 42, 42, 42, 42, 1,),
		('floatint32s', 2147483647, -2147483647, 2147483647, -2147483647, 42,),
		('floatint32si', 2147483647, -2147483647, 2147483647, -2147483647, 1,),
		('floatint64e', 10000000000, -2147483648, 42, -2200000000, 42,),
		('floatint64ei', 10000000000, -2147483648, 42, -2200000000, 1,),
		('floatint64s', 10000000000, -2147483648, 9223372036854775807, -2200000000, 42,),
		('floatint64si', 10000000000, -2147483648, 9223372036854775807, -2200000000, 1,),
	):
		verify(typ, [typ], [b'1.99', b'-3000', b'1e10', b'-2147483648', b'1e100', b'-2.2e9', b'-7.89', b'1.half'], [1, -3000, smallbig, smallneg, bigbig, bigneg, -7, extra], '42')
	def check_special(got, fromstr):
		msg = 'Expected [inf, -inf, nan, nan, nan, nan, inf], got %r from %s.' % (got, fromstr,)
		for ix, v in ((0, float('inf')), (1, float('-inf')), (-1, float('inf'))):
			assert got[ix] == v, msg
		for ix in range(2, 6):
			v = got[ix]
			if isinstance(v, complex):
				v = v.real
			assert isnan(v), msg
	verify('special floats', ['complex32', 'complex64', 'float32', 'float64', 'number'], [b'+Inf', b'-inF', b'nan', b'NaN', b'NAN', b'Nan', b'INF'], check_special)
	verify('complex', ['complex32', 'complex64'], [b'0.25-5j', b'-0.5j', b'infj'], [0.25-5j, -0.5j, complex(0, float('inf'))])
	if options.numeric_comma:
		verify('complex numeric_comma', ['complex32', 'complex64'], [b'-0,5+1,5j', b'-0,5+1.5j', b',5j'], [-0.5+1.5j, 42j, 0.5j], '42j', numeric_comma=True)

def test_bytes():
	want = [b'foo', b'\\bar', b'bl\xe4', b'\x00\t \x08', b'a\xa0b\x255\x00']
	for typs, data in (
		(['bytes', 'bytesstrip'], want,),
		(['bytesstrip'], [b'\t\t\t\rfoo\n \x0c', b'\\bar', b'bl\xe4   \x0b', b' \x00\t \x08', b'a\xa0b\x255\x00\n'],),
	):
		verify(typs[0], typs, data, want)

def test_ascii():
	for prefix, data in (
		('ascii', [b'foo', b'\\bar', b'bl\xe4', b'\x00\t \x08'],),
		('asciistrip', [b'\t\t\t\rfoo\n \x0c', b'\\bar', b'bl\xe4   \x0b', b' \x00\t \x08\t'],),
	):
		for encode, want in (
			(None,      ['foo', '\\bar',    'bl\\344', '\x00\t \x08'],),
			('replace', ['foo', '\\bar',    'bl\\344', '\x00\t \x08'],),
			('encode',  ['foo', '\\134bar', 'bl\\344', '\x00\t \x08'],),
			('strict',  ['foo', '\\bar',    None,      '\x00\t \x08'],),
		):
			if encode:
				typ = prefix + ':' + encode
			else:
				typ = prefix
			verify(typ, [typ], data, want, default=None)
	verify(
		'ascii all source types', ['ascii', 'asciistrip'],
		[b'\t\t\t\rfoo\r \x0c', b'\\bar', b' \x00\t \x08\t'],
		{
			'ascii': ['\t\t\t\rfoo\r \x0c', '\\bar', ' \x00\t \x08\t'],
			'asciistrip': ['foo', '\\bar', '\x00\t \x08'],
		},
		all_source_types=True,
	)

def test_unicode():
	data = [b'foo bar\n', b'\tbl\xe5', b'\tbl\xe5a \xe4  ', b'\tbl\xc3\xa5a ', b'\r+AOU-h']
	want = {
		'unicode:utf-8': ['foo bar\n', '既定 ', '既定 ', '\tblåa ', '\r+AOU-h'],
		'unicode:utf-8/strict': ['foo bar\n', '既定 ', '既定 ', '\tblåa ', '\r+AOU-h'],
		'unicode:utf-8/replace': ['foo bar\n', '\tbl\ufffd', '\tbl\ufffda \ufffd  ', '\tblåa ', '\r+AOU-h'],
		'unicode:utf-8/ignore': ['foo bar\n', '\tbl', '\tbla   ', '\tblåa ', '\r+AOU-h'],
		'unicode:iso-8859-1': ['foo bar\n', '\tblå', '\tblåa ä  ', '\tbl\xc3\xa5a ', '\r+AOU-h'],
		'unicode:ascii/replace': ['foo bar\n', '\tbl\ufffd', '\tbl\ufffda \ufffd  ', '\tbl\ufffd\ufffda ', '\r+AOU-h'],
		'unicode:ascii/ignore': ['foo bar\n', '\tbl', '\tbla   ', '\tbla ', '\r+AOU-h'],
		'unicode:utf-7/ignore': ['foo bar\n', '\tbl', '\tbla   ', '\tbla ', '\råh'],
		'unicode:utf-7/replace': ['foo bar\n', '\tbl\ufffd', '\tbl\ufffda \ufffd  ', '\tbl\ufffd\ufffda ', '\råh'],
		'unicodestrip:utf-8': ['foo bar', '既定', '既定', 'blåa', '+AOU-h'],
		'unicodestrip:utf-8/strict': ['foo bar', '既定', '既定', 'blåa', '+AOU-h'],
		'unicodestrip:utf-8/replace': ['foo bar', 'bl\ufffd', 'bl\ufffda \ufffd', 'blåa', '+AOU-h'],
		# strip happens before ignore, so strip+ignore will have one space from '\tbl\xe5a \xe4  '
		'unicodestrip:utf-8/ignore': ['foo bar', 'bl', 'bla ', 'blåa', '+AOU-h'],
		'unicodestrip:iso-8859-1': ['foo bar', 'blå', 'blåa ä', 'bl\xc3\xa5a', '+AOU-h'],
		'unicodestrip:ascii/replace': ['foo bar', 'bl\ufffd', 'bl\ufffda \ufffd', 'bl\ufffd\ufffda', '+AOU-h'],
		'unicodestrip:ascii/ignore': ['foo bar', 'bl', 'bla ', 'bla', '+AOU-h'],
		'unicodestrip:utf-7/ignore': ['foo bar', 'bl', 'bla ', 'bla', 'åh'],
		'unicodestrip:utf-7/replace': ['foo bar', 'bl\ufffd', 'bl\ufffda \ufffd', 'bl\ufffd\ufffda', 'åh'],
	}
	verify('unicode', list(want), data, want, default='既定 ');
	# 既定 is not ascii, so we need a separate default for ascii-like types.
	want = {
		'unicode:ascii/strict': ['foo bar\n', 'standard', 'standard', 'standard', '\r+AOU-h'],
		'unicodestrip:ascii/strict': ['foo bar', 'standard', 'standard', 'standard', '+AOU-h'],
		'unicodestrip:utf-7/strict': ['foo bar', 'standard', 'standard', 'standard', 'åh'],
	}
	verify('unicode with ascii default', list(want), data, want, default='standard');
	verify('utf7 all', ['unicode:utf-7/replace'], [b'a+b', b'a+-b', b'+ALA-'], ['a\ufffd', 'a+b', '°'], all_source_types=True);

def test_datetimes():
	# These use the libc functions, so may only work for dates after 1900.
	# They also use the python datetime classes to verify valid dates,
	# so on Python < 3.6 they will accept various invalid dates if strptime
	# does. So we only try nearly-good dates on Python 3.6+.
	todo = [
		('date YYYYMMDD', 'date:%Y%m%d', [b'20190521', b'19700101', b'1970-01-01', b'1980', b'nah'], [date(2019, 5, 21), date(1970, 1, 1), date(1945, 6, 20), date(1945, 6, 20), date(1945, 6, 20)], '19450620', True,),
		('date spacy YYYYMMDD', 'date: %Y %m %d', [b'20190521', b'   1970 01\n\n\n01', b'1970\t1 1    ', b'70 0 0', b'1981'], [date(2019, 5, 21), date(1970, 1, 1), date(1970, 1, 1), date(1945, 6, 20), date(1945, 6, 20)], '1945 6 20', False,),
		('date YYYYblahMMDD', 'date:%Yblah%m%d', [b'2019blah0521', b'1970blah0101', b'1970blah-01-01', b'1980blah', b'nah'], [date(2019, 5, 21), date(1970, 1, 1), date(1945, 6, 20), date(1945, 6, 20), date(1945, 6, 20)], '1945blah0620', True,),
		('datetime hhmmYYMMDD', 'datetime:%H%M%y%m%d', [b'1852190521', b'0000700101', b'today'], [datetime(2019, 5, 21, 18, 52), datetime(1970, 1, 1), datetime(1978, 1, 1)], '0000780101', False,),
		('datetime YYYYMMDD HHMMSS.mmmmmm', 'datetime:%Y%m%d %H%M%S.%f', [b'20190521 185206.123', b'19700203040506.000007', b'19700203040506.-00007', b'today'], [datetime(2019, 5, 21, 18, 52, 6, 123000), datetime(1970, 2, 3, 4, 5, 6, 7), datetime(1978, 1, 1), datetime(1978, 1, 1)], '19780101 000000.0', True,),
		('time HH:MM', 'time:%H:%M', [b'03:14', b'18:52', b'25:10'], [time(3, 14), time(18, 52), time(0, 42)], '00:42', True,),
		('time HHMMpercentfSS', 'time:%H%M%%f%S', [b'0314%f00', b'1852%f09', b'1938%f60'], [time(3, 14), time(18, 52, 9), time(0, 42, 18)], '0042%f18', False,),
		('time HHMM mmm.SS', 'time:%H%M %f.%S', [b'03149.00', b'1852456.09', b'1938   123456.44', b'1938123456.60', b'1852456   .09', b'1852456    .09'], [time(3, 14, 0, 900000), time(18, 52, 9, 456000), time(19, 38, 44, 123456), time(0, 42, 18), time(18, 52, 9, 456000), time(0, 42, 18)], '004200.18', False,),
		('date MMDD', 'date:%m%d', [b'0101', b'1020', b'nah'], [date(1970, 1, 1), date(1970, 10, 20), date(1970, 6, 20)], '0620', False,),
		('datetime YY', 'datetime:%y', [b'70', b'2000', b'19'], [datetime(1970, 1, 1, 0, 0, 0), None, datetime(2019, 1, 1, 0, 0, 0)], None, False,),
		('datetime mmmmmmDD', 'datetime:%f%d', [b'00030030', b'00000006', b'00003003', b'99999999', b'99999911'], [datetime(1970, 1, 30, microsecond=300), datetime(1970, 1, 6), datetime(1970, 1, 3, microsecond=30), None, datetime(1970, 1, 11, microsecond=999999)], None, False,),
		('datetime mmmmmm.DD', 'datetime:%f.%d', [b'30.30', b'0.06', b'00030.03', b'999999.99', b'999999.11'], [datetime(1970, 1, 30, microsecond=300000), datetime(1970, 1, 6), datetime(1970, 1, 3, microsecond=300), None, datetime(1970, 1, 11, microsecond=999999)], None, False,),
		('datetime mmmmmmpercentfpercentDD', 'datetime:%f%%f%%%d', [b'30%f%30', b'0%f%06', b'00030%f%03', b'999999%f%99', b'999999%f%11'], [datetime(1970, 1, 30, microsecond=300000), datetime(1970, 1, 6), datetime(1970, 1, 3, microsecond=300), None, datetime(1970, 1, 11, microsecond=999999)], None, False,),
		('datetime unix.f', 'datetime:%s.%f', [b'30.30', b'1558662853.847211', b''], [datetime(1970, 1, 1, 0, 0, 30, 300000), datetime(2019, 5, 24, 1, 54, 13, 847211), datetime(1970, 1, 1, microsecond=100000)], '0.1', False,),
		('datetime java', 'datetime:%J', [b'0', b'1558662853847', b'', b'-2005'], [datetime(1970, 1, 1), datetime(2019, 5, 24, 1, 54, 13, 847000), datetime(1970, 1, 1, 0, 0, 0, 1000), datetime(1969, 12, 31, 23, 59, 57, 995000)], '1', False,),
		('datetime java blahbluh', 'datetime:blah%Jbluh', [b'blah0bluh', b'blah   30000bluh', b'bla0bluh', b'blah0blu', b'blah-2005bluh'], [datetime(1970, 1, 1), datetime(1970, 1, 1, 0, 0, 30), datetime(1970, 1, 1, 0, 0, 0, 1000), datetime(1970, 1, 1, 0, 0, 0, 1000), datetime(1969, 12, 31, 23, 59, 57, 995000)], 'blah1bluh', False,),
	]
	if sys.version_info >= (3, 6):
		todo.extend((
			('nearly good date YYYY-MM-DD', 'date:%Y-%m-%d', [b'2019-02-29', b'1970-02-31', b'1980-06-31', b'1992-02-29'], [None, None, None, date(1992, 2, 29)], None, False,),
			('nearly good datetime YYYY-MM-DD', 'datetime:%Y-%m-%d', [b'2019-02-29', b'1970-02-31', b'1980-06-31', b'1992-02-29'], [None, None, None, datetime(1992, 2, 29)], None, False,),
		))
	for name, typ, data, want, default, all_source_types in todo:
		verify(name, [typ], data, want, default, all_source_types=all_source_types)
		if default is not None:
			if typ.endswith('%f') or typ.endswith('%J'):
				idata = [v + b'abc123' for v in data]
				default += 'a2'
			else:
				idata = [v + b'1868' for v in data]
				default += '42'
			verify(name + ' i', [typ.replace(':', 'i:', 1)], idata, want, default)
	# Timezone tests. I hope all systems accept the :Region/City syntax.
	verify('tz a', ['datetime:%Y-%m-%d %H:%M'], [b'2020-09-30 11:44'], [datetime(2020, 9, 30, 11, 44)])
	verify('tz b', ['datetime:%Y-%m-%d %H:%M'], [b'2020-09-30 11:44'], [datetime(2020, 9, 30, 11, 44)], timezone='UTC')
	verify('tz c', ['datetime:%Y-%m-%d %H:%M'], [b'2020-09-30 13:44', b'2020-02-22 12:44'], [datetime(2020, 9, 30, 11, 44), datetime(2020, 2, 22, 11, 44)], timezone=':Europe/Stockholm')

def test_filter_bad_across_types():
	columns={
		'bytes': 'bytes',
		'float64': 'bytes',
		'int32_10': 'ascii',
		'json': 'unicode',
		'number:int': 'unicode',
		'unicode:utf-8': 'bytes',
	}
	# all_good, *values
	# Make sure all those types (except bytes) can filter other lines,
	# and be filtered by other lines. And that several filtering values
	# is not a problem (line 11).
	data = [
		[True,  b'first',    b'1.1', '1',  '"a"',   '001', b'ett',],
		[True,  b'second',   b'2.2', '2',  '"b"',   '02',  b'tv\xc3\xa5',],
		[True,  b'third',    b'3.3', '3',  '["c"]', '3.0', b'tre',],
		[False, b'fourth',   b'4.4', '4',  '"d"',   '4.4', b'fyra',],       # number:int bad
		[False, b'fifth',    b'5.5', '-',  '"e"',   '5',   b'fem',],        # int32_10 bad
		[False, b'sixth',    b'6.b', '6',  '"f"',   '6',   b'sex',],        # float64 bad
		[False, b'seventh',  b'7.7', '7',  '{"g"}', '7',   b'sju',],        # json bad
		[False, b'eigth',    b'8.8', '8',  '"h"',   '8',   b'\xa5\xc3tta',],# unicode:utf-8 bad
		[True,  b'ninth',    b'9.9', '9',  '"i"',   '9',   b'nio',],
		[True,  b'tenth',    b'10',  '10', '"j"',   '10',  b'tio',],
		[False, b'eleventh', b'11a', '1-', '"k",',  '1,',  b'elva',],       # float64, int32_10 and number:int bad
		[True,  b'twelfth',  b'12',  '12', '"l"',   '12',  b'tolv',],
	]
	want_bad = [tuple(l[1:]) for l in data if not l[0]]
	dw = DatasetWriter(name="filter bad across types", columns=columns, allow_missing_slices=True)
	cols_to_check = ['int32_10', 'bytes', 'json', 'unicode:utf-8']
	if PY3:
		# z so it sorts last.
		dw.add('zpickle', 'pickle')
		cols_to_check.append('zpickle')
		for ix in range(len(data)):
			data[ix].append({ix})
	dw.set_slice(0)
	want = []
	def add_want(ix):
		v = data[ix]
		want.append((int(v[3]), v[1], json.loads(v[4]), v[6].decode('utf-8'),))
		if PY3:
			want[-1] = want[-1] + (v[7],)
	for ix, v in enumerate(data):
		if v[0]:
			add_want(ix)
		dw.write(*v[1:])
	source_ds = dw.finish()
	# Once with just filter_bad, once with some defaults too.
	defaults = {}
	for _ in range(2):
		jid = subjobs.build(
			'dataset_type',
			datasets=dict(source=source_ds),
			options=dict(column2type={t: t for t in columns}, filter_bad=True, defaults=defaults),
		)
		typed_ds = Dataset(jid)
		got = list(typed_ds.iterate(0, cols_to_check))
		assert got == want, "Expected %r, got %r from %s (from %r%s)" % (want, got, typed_ds, source_ds, ' with defaults' if defaults else '')
		bad_ds = Dataset(jid, 'bad')
		got_bad = list(bad_ds.iterate(0, sorted(columns)))
		assert got_bad == want_bad, "Expected %r, got %r from %s (from %r%s)" % (want_bad, got_bad, bad_ds, source_ds, ' with defaults' if defaults else '')
		# make more lines "ok" for the second lap
		if not defaults:
			want_bad.pop(0) # number:int
			want_bad.pop(1) # float64
			want_bad.pop(1) # json
		defaults = {'number:int': '0', 'float64': '0', 'json': '"replacement"'}
		add_want(3)
		add_want(5)
		data[6][4] = '"replacement"'
		add_want(6)
		want.sort() # adding them out of order, int32_10 sorts correctly.

def test_filter_bad_with_rename_and_chain():
	dw = DatasetWriter(name="filter bad with rename", allow_missing_slices=True)
	dw.add('a', 'ascii')
	dw.add('b', 'bytes')
	dw.add('c', 'unicode')
	dw.set_slice(0)
	dw.write('0', b'1', '2')
	dw.write('9', B'A', 'B')
	dw.write('C', B'D', 'E')
	source_ds = dw.finish()
	jid = subjobs.build(
		'dataset_type',
		column2type=dict(b='int32_10', c='int64_16', d='int32_16'),
		filter_bad=True,
		rename=dict(a='b', b='c', c='d'),
		source=source_ds,
	)
	typed_ds = jid.dataset()
	coltypes = sorted((name, col.type) for name, col in typed_ds.columns.items())
	assert coltypes == [('b', 'int32'), ('c', 'int64'), ('d', 'int32')], coltypes
	assert list(typed_ds.iterate(0)) == [(0, 1, 2), (9, 10, 11)]
	bad_ds = jid.dataset('bad')
	coltypes = sorted((name, col.type) for name, col in bad_ds.columns.items())
	assert coltypes == [('b', 'ascii'), ('c', 'bytes'), ('d', 'unicode')], coltypes
	assert list(bad_ds.iterate(0)) == [('C', b'D', 'E')]

	dw = DatasetWriter(name="filter bad with rename chain", allow_missing_slices=True, previous=source_ds)
	dw.add('a', 'ascii')
	dw.add('b', 'ascii')
	dw.add('c', 'ascii')
	dw.set_slice(0)
	dw.write('3', '4', '5')
	dw.write('6', '7', 'eight')
	source_ds = dw.finish()
	jid = subjobs.build(
		'dataset_type',
		column2type=dict(a='number', b='int32_10', c='int64_10'),
		defaults=dict(a='8'),
		filter_bad=True,
		rename=dict(a='b', b='c', c='a'),
		source=source_ds,
	)
	typed_ds = jid.dataset()
	coltypes = sorted((name, col.type) for name, col in typed_ds.columns.items())
	assert coltypes == [('a', 'number'), ('b', 'int32'), ('c', 'int64')], coltypes
	assert list(typed_ds.iterate(0)) == [(2, 0, 1), (5, 3, 4), (8, 6, 7)]
	bad_ds = jid.dataset('bad')
	coltypes = sorted((name, col.type) for name, col in bad_ds.columns.items())
	assert coltypes == [('a', 'unicode'), ('b', 'ascii'), ('c', 'bytes')], coltypes
	assert list(bad_ds.iterate(0)) == [('B', '9', b'A'), ('E', 'C', b'D')]

def test_column_discarding():
	dw = DatasetWriter(name='column discarding')
	dw.add('a', 'bytes')
	dw.add('b', 'bytes')
	dw.add('c', 'bytes')
	w = dw.get_split_write()
	w(b'a', b'b', b'c')
	source = dw.finish()

	# Discard b because it's not typed
	ac_implicit = subjobs.build(
		'dataset_type',
		source=source,
		column2type=dict(a='ascii', c='ascii'),
		discard_untyped=True,
	).dataset()
	assert sorted(ac_implicit.columns) == ['a', 'c'], '%s: %r' % (ac_implicit, sorted(ac_implicit.columns),)
	assert list(ac_implicit.iterate(None)) == [('a', 'c',)], ac_implicit

	# Discard b explicitly
	ac_explicit = subjobs.build(
		'dataset_type',
		source=source,
		column2type=dict(a='ascii', c='ascii'),
		rename=dict(b=None),
	).dataset()
	assert sorted(ac_explicit.columns) == ['a', 'c'], '%s: %r' % (ac_explicit, sorted(ac_explicit.columns),)
	assert list(ac_explicit.iterate(None)) == [('a', 'c',)], ac_explicit

	# Discard c by overwriting it with b. Keep untyped b.
	ac_bASc = subjobs.build(
		'dataset_type',
		source=source,
		column2type=dict(a='ascii', c='ascii'),
		rename=dict(b='c'),
	).dataset()
	assert sorted(ac_bASc.columns) == ['a', 'b', 'c'], '%s: %r' % (ac_bASc, sorted(ac_bASc.columns),)
	assert list(ac_bASc.iterate(None)) == [('a', b'b', 'b',)], ac_bASc

	# Discard c by overwriting it with b. Also type b as a different type.
	abc_bASc = subjobs.build(
		'dataset_type',
		source=source,
		column2type=dict(a='ascii', b='strbool', c='ascii'),
		rename=dict(b='c'),
	).dataset()
	assert sorted(abc_bASc.columns) == ['a', 'b', 'c'], '%s: %r' % (abc_bASc, sorted(abc_bASc.columns),)
	assert list(abc_bASc.iterate(None)) == [('a', True, 'b',)], abc_bASc

def test_rehash_with_empty_slices():
	dw = DatasetWriter(name='rehash with empty slices', hashlabel='a')
	dw.add('a', 'ascii')
	dw.add('b', 'ascii')
	w = dw.get_split_write()
	w('a', '42')
	w('42', 'b')
	source = dw.finish()
	hashfunc = typed_writer('int32').hash
	def verify_hashing(caption, want_values, **kw):
		ds = subjobs.build(
			'dataset_type',
			source=source,
			column2type=dict(a='int32_10'),
			caption=caption,
			**kw
		).dataset()
		got_values = set()
		for sliceno in range(g.slices):
			for got in ds.iterate(sliceno):
				assert hashfunc(got[0]) % g.slices == sliceno
				assert got not in got_values
				got_values.add(got)
		assert want_values == got_values
	verify_hashing('with discard', {(42, 'b',)}, filter_bad=True)
	# using defaults uses some different code paths
	verify_hashing('with default=0 (probably two slices)', {(0, '42',), (42, 'b',)}, defaults=dict(a='0'))
	verify_hashing('with default=42 (one slice)', {(42, '42',), (42, 'b',)}, defaults=dict(a='42'))

def synthesis():
	test_bytes()
	test_ascii()
	test_unicode()
	test_numbers()
	test_datetimes()
	test_column_discarding()
	test_rehash_with_empty_slices()

	verify('json', ['json'],
		[b'null', b'[42, {"a": "b"}]', b'\r  {  "foo":\r"bar" \r   }\t ', b'nope'],
		[None, [42, {'a': 'b'}], {'foo': 'bar'}, ['nah']],
		default='["nah"]', all_source_types=True,
	)

	test_filter_bad_across_types()
	test_filter_bad_with_rename_and_chain()

	for t in (all_typenames - used_typenames):
		print(t)



############################################################################
#                                                                          #
# Copyright (c) 2019-2021 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test that hashlabel does what it says in both split_write and hashcheck.
Then test that rehashing gives the expected result, and that using the
wrong hashlabel without rehashing is not allowed.
'''

from accelerator.dataset import DatasetWriter
from accelerator.extras import DotDict
from accelerator.dsutil import typed_writer
from accelerator.error import DatasetUsageError
from accelerator.compat import unicode

from datetime import datetime, date

all_data = list(zip(range(10000), reversed(range(10000))))

def prepare(params):
	assert params.slices >= 2, "Hashing won't do anything with just one slice"
	dws = DotDict()
	# all the numeric types should hash the same (for values they have in common)
	for name, hashlabel, typ in (
		("unhashed_manual"    , None  , "int32"),     # manually interlaved
		("unhashed_split"     , None  , "int64"),     # split_write interlaved
		("up_checked"         , "up"  , "float32"),   # hashed on up using dw.hashcheck
		("up_split"           , "up"  , "float64"),   # hashed on up using split_write
		("down_checked"       , "down", "bits32"),    # hashed on down using dw.hashcheck
		("down_discarded"     , "down", "bits64"),    # hashed on down using discarding writes
		("down_discarded_list", "down", "number"),    # hashed on down using discarding list writes
		("down_discarded_dict", "down", "complex32"), # hashed on down using discarding dict writes
		# we have too many types, so we need more datasets
		("unhashed_complex64" , None  , "complex64"),
		("unhashed_bytes"     , None  , "bytes"),
		("up_ascii"           , "up"  , "ascii"),
		("down_unicode"       , "down", "unicode"),
		# datetime on 1970-01-01 hashes like time
		("up_datetime"        , "up"  , "datetime"),
		("down_time"          , "down", "time"),
		# date doesn't hash the same as anything else, so compare it to itself
		("up_date"            , "up"  , "date"),
		("down_date"          , "down", "date"),
	):
		dw = DatasetWriter(name=name, hashlabel=hashlabel)
		dw.add("up", typ)
		dw.add("down", typ)
		dws[name] = dw
	return dws

def analysis(sliceno, prepare_res, params):
	dws = prepare_res
	dws.down_discarded.enable_hash_discard()
	dws.down_discarded_list.enable_hash_discard()
	dws.down_discarded_dict.enable_hash_discard()
	dws.up_datetime.enable_hash_discard()
	dws.down_time.enable_hash_discard()
	dws.up_ascii.enable_hash_discard()
	dws.down_unicode.enable_hash_discard()
	dws.up_date.enable_hash_discard()
	dws.down_date.enable_hash_discard()
	for ix, (up, down) in enumerate(all_data):
		if dws.up_checked.hashcheck(up):
			dws.up_checked.write(up, down)
		if dws.down_checked.hashcheck(down):
			dws.down_checked.write(up, down)
		if ix % params.slices == sliceno:
			dws.unhashed_manual.write(up, down)
			dws.unhashed_complex64.write(up, down)
			dws.unhashed_bytes.write(str(up).encode("ascii"), str(down).encode("ascii"))
		dws.down_discarded.write(up, down)
		dws.down_discarded_list.write_list([up, down])
		dws.down_discarded_dict.write_dict(dict(up=up, down=down))
		dt_up = datetime(1970, 1, 1, 0, 0, 0, up)
		dt_down = datetime(1970, 1, 1, 0, 0, 0, down)
		dws.up_datetime.write(dt_up, dt_down)
		dws.down_time.write(dt_up.time(), dt_down.time())
		dws.up_date.write(date.fromordinal(up + 1), date.fromordinal(down + 1))
		dws.down_date.write(date.fromordinal(up + 1), date.fromordinal(down + 1))
		dws.up_ascii.write(str(up), str(down))
		dws.down_unicode.write(unicode(up), unicode(down))
	# verify that we are not allowed to write in the wrong slice without enable_hash_discard
	if not dws.up_checked.hashcheck(0):
		good = True
		for fn, a in (
			("write", (0, 0,)),
			("write_list", ([0, 0],)),
			("write_dict", (dict(up=0, down=0),)),
		):
			try:
				getattr(dws.up_checked, fn)(*a)
				good = False
			except Exception:
				pass
			assert good, "%s allowed writing in wrong slice" % (fn,)

# complex isn't sortable
def uncomplex(t):
	if isinstance(t[0], complex):
		return tuple(v.real for v in t)
	else:
		return t

# datetime doesn't compare to time (or numbers, for the all_data check)
def undatetime(t):
	if hasattr(t[0], "microsecond"):
		return tuple(v.microsecond for v in t)
	elif hasattr(t[0], "toordinal"):
		return tuple(v.toordinal() - 1 for v in t)
	else:
		return t

def unstr(t):
	if isinstance(t[0], (bytes, unicode)):
		return tuple(int(v) for v in t)
	else:
		return t

def cleanup(lst):
	return [unstr(undatetime(uncomplex(t))) for t in lst]

def synthesis(prepare_res, params, job, slices):
	dws = prepare_res
	for dw in (dws.unhashed_split, dws.up_split,):
		w = dw.get_split_write_list()
		for row in all_data:
			w(row)
	hl2ds = {None: [], "up": [], "down": []}
	all_ds = {}
	special_cases = {
		"up_datetime", "down_time", "up_date", "down_date",
		"unhashed_bytes", "up_ascii", "down_unicode",
	}
	for name, dw in dws.items():
		ds = dw.finish()
		all_ds[ds.name] = ds
		if ds.name not in special_cases:
			hl2ds[ds.hashlabel].append(ds)

	# Verify that the different ways of writing gave the same result
	for hashlabel in (None, "up", "down"):
		for sliceno in range(slices):
			data = [(ds, list(ds.iterate(sliceno))) for ds in hl2ds[hashlabel]]
			good = data[0][1]
			for name, got in data:
				assert got == good, "%s doesn't match %s in slice %d" % (data[0][0], name, sliceno,)

	# Verify that both up and down hashed on the expected column
	hash = typed_writer("int32").hash
	for colname in ("up", "down"):
		ds = all_ds[colname + "_checked"]
		for sliceno in range(slices):
			for value in ds.iterate(sliceno, colname):
				assert hash(int(value)) % slices == sliceno, "Bad hashing on %s in slice %d" % (colname, sliceno,)

	# Verify that up and down are not the same, to catch hashing
	# not actually hashing.
	for up_name, down_name in (
		("up_checked", "down_checked"),
		("up_datetime", "down_time"),
		("up_date", "down_date"),
		("up_ascii", "down_unicode"),
	):
		up = cleanup(all_ds[up_name].iterate(None))
		down = cleanup(all_ds[down_name].iterate(None))
		assert up != down, "Hashlabel did not change slice distribution (%s vs %s)" % (up_name, down_name)
		# And check that the data is still the same.
		assert sorted(up) == sorted(down) == all_data, "Hashed datasets have wrong data (%s vs %s)" % (up_name, down_name)

	# Verify that rehashing works.
	# (Can't use sliceno None, because that won't rehash, and even if it did
	# the order wouldn't match. Order doesn't even match in the rehashed
	# individual slices.)
	def test_rehash(want_ds, chk_ds_lst):
		want_ds = all_ds[want_ds]
		for sliceno in range(slices):
			want = sorted(cleanup(want_ds.iterate(sliceno)))
			for chk_ds in chk_ds_lst:
				assert chk_ds.hashlabel != want_ds.hashlabel
				got = chk_ds.iterate(sliceno, hashlabel=want_ds.hashlabel, rehash=True)
				got = sorted(cleanup(got))
				assert want == got, "Rehashing is broken for %s (slice %d of %s)" % (chk_ds.columns[want_ds.hashlabel].type, sliceno, chk_ds,)
	test_rehash("up_checked", hl2ds[None] + hl2ds["down"])
	test_rehash("down_checked", hl2ds[None] + hl2ds["up"])
	test_rehash("up_datetime", [all_ds["down_time"]])
	test_rehash("down_time", [all_ds["up_datetime"]])
	test_rehash("down_date", [all_ds["up_date"]])
	test_rehash("up_ascii", [all_ds["unhashed_bytes"], all_ds["down_unicode"]])
	test_rehash("down_unicode", [all_ds["unhashed_bytes"], all_ds["up_ascii"]])

	# And finally verify that we are not allowed to specify the wrong hashlabel
	good = True
	try:
		all_ds["up_checked"].iterate(None, hashlabel="down")
		good = False
	except DatasetUsageError:
		pass
	try:
		all_ds["unhashed_manual"].iterate(None, hashlabel="down")
		good = False
	except DatasetUsageError:
		pass
	assert good, "Iteration allowed on the wrong hashlabel"

	# verify that non-integral floats hash the same in the five types that can have them
	# using + 0.5 is safe for the values we use, it can be exactly represented in 32 bit floats.
	float_data = [v + 0.5 for v, _ in all_data]
	float_ds_lst = []
	for typ in ("float32", "float64", "complex32", "complex64", "number"):
		dw = job.datasetwriter(name="floattest_" + typ, columns={"value": typ}, hashlabel="value")
		write = dw.get_split_write()
		for v in float_data:
			write(v)
		float_ds_lst.append(dw.finish())
	for sliceno in range(slices):
		values = [(ds, list(ds.iterate(sliceno, "value"))) for ds in float_ds_lst]
		want_ds, want = values.pop()
		for ds, got in values:
			assert got == want, "%s did not match %s in slice %d" % (ds, want_ds, sliceno,)



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from . import a_test_output

description = r'''
Run from test_output to check that output handling works correctly
when only prepare and analysis are defined.
'''

depend_extra = (a_test_output,)

options = dict(
	prefix='',
	p='',
	a='',
)

def prepare():
	a_test_output.sub_part('p', options)

def analysis(sliceno):
	a_test_output.sub_part(sliceno, options)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify that some potentially problematic dataset names work.
'''

def synthesis(job):
	def mk(name):
		dw = job.datasetwriter(name=name, columns={'name': 'unicode'})
		dw.get_split_write()(name)
		ds = dw.finish()
		assert ds.name == name
		return job.dataset(name)
	names = [
		'',            # empty is not a valid fs name
		'.',           # . is a reserved fs name
		'..',          # .. is a reserved fs name
		'\0',          # many things have problems with \0 ("string terminator")
		'\n',          # newlines are difficult
		' ',           # sometimes spaces are too
		'\xa0',        # this is non-breaking space
		'dataset.txt', # name collides with the legacy list of dataset names
		'LIST',        # name collides with the pickle of dataset names
		'ᛞᚨᛏᚨᛊᛖᛏ',     # non-ascii
		'😁',          # this character doesn't fit in 16 bits
	]
	for name in names:
		assert list(mk(name).iterate(None, 'name')) == [name], name
	return names



from .printer import prt

description = "Dataset: Iterate"

def main(urd):
	prt.source(__file__)
	prt()
	prt('Run a method that creates a dataset.')
	job = urd.build('dsexample_createdataset')

	prt()
	prt('Iterate and merge the number column of the dataset.')
	job = urd.build('dsexample_iterateandmerge', source=job)
	prt()
	prt("The job's return value is:", job.load())



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test sliceno="roundrobin" in dataset iteration.
'''

from accelerator.dataset import DatasetWriter

def prepare(params):
	assert params.slices >= 3
	dw_3 = DatasetWriter(name="three", columns={"num": "int32"})
	dw_long = DatasetWriter(name="long", columns={"num": "int32"})
	dw_uneven = DatasetWriter(name="uneven", columns={"num": "int32"})
	return dw_3, dw_long, dw_uneven

def analysis(sliceno, prepare_res, params):
	dw_3, dw_long, dw_uneven = prepare_res
	if sliceno < 3:
		dw_3.write(sliceno)
		dw_3.write(sliceno + 3)
	for ix in range(sliceno, 100000, params.slices):
		dw_long.write(ix)
	dw_uneven.write(sliceno)
	if sliceno > 1:
		dw_uneven.write(100000 + sliceno)
	if sliceno == 2:
		dw_uneven.write(-1)
		dw_uneven.write(-2)

def synthesis(prepare_res, params):
	dw_3, dw_long, dw_uneven = prepare_res
	ds_3 = dw_3.finish()
	ds_long = dw_long.finish()
	ds_uneven = dw_uneven.finish()
	inorder = list(ds_3.iterate(None, "num"))
	assert inorder == [0, 3, 1, 4, 2, 5]
	rr = list(ds_3.iterate("roundrobin", "num"))
	assert rr == [0, 1, 2, 3, 4, 5]
	rr = list(ds_long.iterate("roundrobin", "num"))
	assert rr == list(range(100000))
	rr = list(ds_uneven.iterate("roundrobin", "num"))
	assert rr == list(range(params.slices)) + list(range(100002, 100000 + params.slices)) + [-1, -2]



from os.path import dirname, join
from .printer import prt

description = "Tutorial: Import, type, sort, and hash partition a file."


# File is stored in same directory as this python file.
# In a real project, set "input directory" in config file instead!
filename = join(dirname(__file__), 'data.csv')


def main(urd):
	prt.source(__file__)
	prt()
	prt('Import a CSV-file.  Type, sort, and hash-partition the imported dataset.')
	prt()

	imp = urd.build('csvimport', filename=filename)
	imp = urd.build(
		'dataset_type',
		source=imp,
		column2type=dict(
			Date='date:%Y-%m-%d',
			String='unicode:utf-8',
			Int='number',
			Float='float64',
		),
	)
	imp = urd.build('dataset_sort', source=imp, sort_columns='Date')
	imp = urd.build('dataset_hashpart', source=imp, hashlabel='String')

	prt()
	prt('Note how the output from a build()-call is used as input')
	prt('to the next in order to pass data and/or parameters and create')
	prt('dependencies.')

	prt()
	with prt.header('VISUALISING A JOB\'S DEPENDENCIES'):
		prt('View the job\'s metadata using')
		prt.command('ax job %s' % (imp,))
		prt('We can see that the dataset "%s" is input to this job.' % (imp.params.datasets.source,))

	prt()
	with prt.header('REFERENCES TO ALL JOBS ARE STORED IN THE "urd.joblist" OBJECT:'):
		prt.output(urd.joblist.pretty)
	prt('All jobs are stored in "urd.joblist" so that they can be easily')
	prt('fetched at a later time.  This will be used in the next step.')

	prt()
	with prt.header('HASH PARTITIONED DATASET'):
		prt('''
			Hash partitioning is a very efficient way to prepare
			a dataset for parallel processing (see manual).

			Take a look at the dataset created by dataset_hashpart:
		''')
		prt.command('ax ds %s' % (imp,))
		prt('''
			The asterisk on the row corresponding to the "String" column
			indicates that the dataset is hash partitioned based the values
			in this column.  It is possible to use the "-s" option to
			"ax cat" to print data from individual slices.)

			We can also see how many rows there are in each slice by typing
		''')
		prt.command('ax ds -s %s' % (imp,))
